from collections.abc import Callable
from uuid import UUID

from pydantic import BaseModel, Field

from gen_epix.casedb.domain.enum import CaseRight, CaseRightSet
from gen_epix.casedb.domain.model.case.non_persistable import CaseRights, CaseSetRights
from gen_epix.fastapp import exc


class CaseTypeAccessAbac(BaseModel):
    case_type_id: UUID = Field(description="The ID of the case type")
    data_collection_id: UUID = Field(description="The ID of the data collection")
    is_private: bool = Field(
        description="Whether the data collection is private, limited to the case types in the case type set. When true, add/remove case and add/remove case set are considered (i) as the right to create/delete a case or case set in this data collection (setting case.created_in_data_collection to this data collection) and (ii) as the right to share the case or case set further in other data collections. Deleting a case or case set is only allowed when it can or has been removed from all other data collections as well."
    )
    add_case: bool = Field(
        description="Whether cases may be added, limited to the case type and data collection"
    )
    remove_case: bool = Field(
        description="Whether cases may be removed, limited to the case type and data collection"
    )
    add_case_set: bool = Field(
        description="Whether case sets may be added, limited to the case type and data collection"
    )
    remove_case_set: bool = Field(
        description="Whether case sets may be removed, limited to the case type and data collection"
    )
    read_case_type_col_ids: set[UUID] = Field(
        description="The IDs of the case type columns for which values can be read, limited to the case type and data collection"
    )
    write_case_type_col_ids: set[UUID] = Field(
        description="The IDs of the case type columns for which values can be updated, limited to the case types in the case type set"
    )
    read_case_set: bool = Field(
        description="Whether case set be read, limited to the case type and data collection"
    )
    write_case_set: bool = Field(
        description="Whether case set be updated, limited to the case type and data collection"
    )

    def has_any_rights(self) -> bool:
        """
        Determine if at least one of the rights is set.
        """
        return (
            self.add_case
            or self.remove_case
            or self.add_case_set
            or self.remove_case_set
            or len(self.read_case_type_col_ids) > 0
            or len(self.write_case_type_col_ids) > 0
            or self.read_case_set
            or self.write_case_set
        )


class CaseTypeShareAbac(BaseModel):
    case_type_id: UUID = Field(description="The ID of the case type")
    data_collection_id: UUID = Field(description="The ID of the data collection")
    add_case_from_data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections from which cases may be added to this data collection, limited to the case type"
    )
    remove_case_from_data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections from which cases may be removed from this data collection, limited to the case type"
    )
    add_case_set_from_data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections from which case sets may be added to this data collection, limited to the case type"
    )
    remove_case_set_from_data_collection_ids: set[UUID] = Field(
        description="The IDs of the data collections from which case sets may be removed from this data collection, limited to the case type"
    )

    def has_any_rights(self) -> bool:
        """
        Determine if at least one of the rights is set.
        """
        return (
            len(self.add_case_from_data_collection_ids) > 0
            or len(self.remove_case_from_data_collection_ids) > 0
            or len(self.add_case_set_from_data_collection_ids) > 0
            or len(self.remove_case_set_from_data_collection_ids) > 0
        )


class CaseAbac(BaseModel):
    is_full_access: bool = Field(
        description="Whether the user has full access, i.e. is not limited by any ABAC policies. If so, the other fields are empty and are to be ignored."
    )
    case_type_access_abacs: dict[UUID, dict[UUID, CaseTypeAccessAbac]] = Field(
        description="The case type data collection ABACs for the user, keyed by case type set ID and then data collection ID"
    )
    case_type_share_abacs: dict[UUID, dict[UUID, CaseTypeShareAbac]] = Field(
        description="The case type share ABACs for the user, keyed by case type set ID and then data collection ID"
    )

    def get_combinations_with_any_rights(self) -> dict[UUID, set[UUID]]:
        """
        Get the dict[case_type_id, set[data_collection_ids]] combinations for which
        there is any access or share right. The sets are guaranteed to be non-empty.
        """
        retval: dict[UUID, set[UUID]] = {}
        for (
            case_type_id,
            data_collection_access_abac_map,
        ) in self.case_type_access_abacs.items():
            data_collection_ids = {
                x
                for x, y in data_collection_access_abac_map.items()
                if y.has_any_rights()
            }
            if data_collection_ids:
                retval[case_type_id] = data_collection_ids
        for (
            case_type_id,
            data_collection_share_abac_map,
        ) in self.case_type_share_abacs.items():
            data_collection_ids = {
                x
                for x, y in data_collection_share_abac_map.items()
                if y.has_any_rights()
            }
            if not data_collection_ids:
                continue
            if case_type_id in retval:
                # Merge with existing data collection IDs
                data_collection_ids = retval[case_type_id].union(data_collection_ids)
            else:
                retval[case_type_id] = data_collection_ids
        return retval

    def get_case_types_with_any_rights(self) -> set[UUID]:
        """
        Get the set[case_type_id] for which there is any access or share right in at
        least one of the data collections.
        """
        retval = set()
        for (
            case_type_id,
            data_collection_access_abac_map,
        ) in self.case_type_access_abacs.items():
            has_right = any(
                x.has_any_rights() for x in data_collection_access_abac_map.values()
            )
            if has_right:
                retval.add(case_type_id)
        for (
            case_type_id,
            data_collection_share_abac_map,
        ) in self.case_type_share_abacs.items():
            has_right = any(
                x.has_any_rights() for x in data_collection_share_abac_map.values()
            )
            if has_right:
                retval.add(case_type_id)
        return retval

    def get_combinations_with_access_right(
        self,
        right: CaseRight,
    ) -> dict[UUID, set[UUID]]:
        """
        Get the dict[case_type_id, set[data_collection_ids]] combinations for which
        there is the given right. The sets are guaranteed to be non-empty.
        """
        retval = {}
        has_right_fn = self._get_has_right_function(right)
        for case_type_id, data in self.case_type_access_abacs.items():
            data_collection_ids = {x for x, y in data.items() if has_right_fn(y)}
            if data_collection_ids:
                retval[case_type_id] = data_collection_ids
        return retval

    def get_case_types_with_access_right(self, right: CaseRight) -> set[UUID]:
        """
        Get the set[case_type_id] for which there is the given right in at least one of
        the data collections.
        """
        retval = set()
        has_right_fn = self._get_has_right_function(right)
        for case_type_id, data in self.case_type_access_abacs.items():
            has_right = any(has_right_fn(x) for x in data.values())
            if has_right:
                retval.add(case_type_id)
        return retval

    def get_case_type_cols_with_any_rights(self) -> set[UUID]:
        """
        Get the set[case_type_col_id] for which there is any read or write access in at
        least one of the data collections.
        """
        retval = set()
        for case_type_id, data in self.case_type_access_abacs.items():
            for data_collection_id, access_abac in data.items():
                retval.update(access_abac.read_case_type_col_ids)
                retval.update(access_abac.write_case_type_col_ids)
        return retval

    def get_data_collections_with_any_rights(self) -> set[UUID]:
        data_collection_ids = set()
        for data_collection_access_abac_map in self.case_type_access_abacs.values():
            for (
                data_collection_id,
                access_abac,
            ) in data_collection_access_abac_map.items():
                if access_abac.has_any_rights():
                    data_collection_ids.add(data_collection_id)
        for data_collection_share_abac_map in self.case_type_share_abacs.values():
            for (
                data_collection_id,
                share_abac,
            ) in data_collection_share_abac_map.items():
                if not share_abac.has_any_rights():
                    continue
                data_collection_ids.add(data_collection_id)
                data_collection_ids.update(share_abac.add_case_from_data_collection_ids)
                data_collection_ids.update(
                    share_abac.remove_case_from_data_collection_ids
                )
                data_collection_ids.update(
                    share_abac.add_case_set_from_data_collection_ids
                )
                data_collection_ids.update(
                    share_abac.remove_case_set_from_data_collection_ids
                )
        return data_collection_ids

    def get_data_collections_with_access_right_for_case_type_col(
        self, case_type_col_id: UUID, right: CaseRight
    ) -> set[UUID]:
        """
        Get the set[data_collection_id] for which there is the given right on the given case_type_col.
        """
        if right == CaseRight.READ_CASE:
            return {
                data_collection_id
                for access_by_data_collection in self.case_type_access_abacs.values()
                for data_collection_id, access_abac in access_by_data_collection.items()
                if case_type_col_id in access_abac.read_case_type_col_ids
            }
        elif right == CaseRight.WRITE_CASE:
            return {
                data_collection_id
                for access_by_data_collection in self.case_type_access_abacs.values()
                for data_collection_id, access_abac in access_by_data_collection.items()
                if case_type_col_id in access_abac.write_case_type_col_ids
            }
        else:
            raise exc.InvalidArgumentsError(
                f"Right {right.value} is invalid for case_type_col access"
            )

    def is_allowed(
        self,
        case_type_id: UUID,
        right: CaseRight,
        is_create_or_delete: bool = False,
        created_in_data_collection_id: UUID | None = None,
        current_data_collection_ids: set[UUID] | None = None,
        tgt_data_collection_ids: set[UUID] | None = None,
    ) -> bool:
        """
        Determine if the given right is allowed for the given case type and data
        collections. This covers the following rights:
        - ADD_CASE/SET: create a case or case set and/or add it to all the given data
          collections
        - REMOVE_CASE/SET: delete a case or case set and/or remove it from all the
          given data collections
        - READ/WRITE_CASE/SET: read a case or case set from at least one of the given
          data collections
        """
        # Special case: full access
        if self.is_full_access:
            return True

        # Parse input
        if current_data_collection_ids is None:
            current_data_collection_ids = set()
        if tgt_data_collection_ids is None:
            tgt_data_collection_ids = set()

        # Get rights for the case type
        access_abac = self.case_type_access_abacs.get(case_type_id, {})
        share_abac = self.case_type_share_abacs.get(case_type_id, {})
        if access_abac is None and share_abac is None:
            # No rights to this case type
            return False

        # Handle each right
        has_right_fn = self._get_has_right_function(right)
        if right in CaseRightSet.ADD.value:
            # Check if the case or case set can be added to all the target data collections
            if is_create_or_delete:
                # Check if the case or case set can be created
                if created_in_data_collection_id is None:
                    raise exc.InvalidArgumentsError(
                        f"created_in_data_collection_id must be provided for right {right.value} if is_create_or_delete=True"
                    )
                if created_in_data_collection_id not in access_abac:
                    # No access to this data collection
                    return False
                if not access_abac[created_in_data_collection_id].is_private:
                    # created_in_data_collection_id is not a private data collection
                    return False
                if current_data_collection_ids:
                    raise exc.InvalidArgumentsError(
                        f"current_data_collection_ids must be empty for right {right.value} if is_create_or_delete=True"
                    )
            # Check for each of the remaining target data collections if the user has
            # the right to add cases or case sets to it
            remaining_data_collection_ids = (
                set()
                if tgt_data_collection_ids is None
                else set(tgt_data_collection_ids)
            )
            remaining_data_collection_ids.discard(
                created_in_data_collection_id  # type:ignore[arg-type]
            )
            if current_data_collection_ids:
                remaining_data_collection_ids = (
                    remaining_data_collection_ids - current_data_collection_ids
                )
            current_data_collection_ids.add(
                created_in_data_collection_id  # type:ignore[arg-type]
            )
            get_share_from_data_collections_fn = (
                self._get_get_share_from_data_collections_function(right)
            )
            for data_collection_id in remaining_data_collection_ids:
                if data_collection_id not in access_abac:
                    # No access to this data collection
                    return False
                if access_abac[data_collection_id].is_private:
                    # Private data collection different from the created in data collection
                    return False
                if not has_right_fn(access_abac[data_collection_id]):
                    # No access right in this data collection -> check share rights
                    if (
                        share_abac is None
                        or data_collection_id not in share_abac
                        or not current_data_collection_ids.intersection(
                            get_share_from_data_collections_fn(
                                share_abac[data_collection_id]
                            )
                        )
                    ):
                        # No direct share rights either
                        # TODO: Check indirect share rights from the provided data collections
                        return False
        elif right in CaseRightSet.REMOVE.value:
            # Check if the case or case set can be deleted from all the target data collections
            if is_create_or_delete:
                # Check if the case or case set can be deleted
                if created_in_data_collection_id is None:
                    raise exc.InvalidArgumentsError(
                        f"created_in_data_collection_id must be provided for right {right.value} if is_create_or_delete=True"
                    )
                if created_in_data_collection_id not in access_abac:
                    # No access to this data collection
                    return False
                if not access_abac[created_in_data_collection_id].is_private:
                    # created_in_data_collection_id is not a private data collection
                    return False
                if tgt_data_collection_ids:
                    raise exc.InvalidArgumentsError(
                        f"tgt_data_collection_ids must be empty for right {right.value} if is_create_or_delete=True"
                    )
                tgt_data_collection_ids = current_data_collection_ids
            if not tgt_data_collection_ids.issubset(current_data_collection_ids):
                raise exc.InvalidArgumentsError(
                    f"tgt_data_collection_ids must be a subset of current_data_collection_ids for right {right.value}"
                )
            # Check for each of the remaining target data collections if the user has
            # the right to remove cases or case sets from it
            remaining_data_collection_ids = set(tgt_data_collection_ids)
            remaining_data_collection_ids.discard(
                created_in_data_collection_id  # type:ignore[arg-type]
            )
            get_share_from_data_collections_fn = (
                self._get_get_share_from_data_collections_function(right)
            )
            for data_collection_id in remaining_data_collection_ids:
                if data_collection_id not in access_abac:
                    # No access to this data collection
                    return False
                if access_abac[data_collection_id].is_private:
                    # Private data collection different from the created in data collection
                    return False
                if not has_right_fn(access_abac[data_collection_id]):
                    # No access right in this data collection -> check share rights
                    if (
                        share_abac is None
                        or data_collection_id not in share_abac
                        or not current_data_collection_ids.intersection(
                            get_share_from_data_collections_fn(
                                share_abac[data_collection_id]
                            )
                        )
                    ):
                        # No direct share rights either
                        # TODO: Check indirect share rights from the provided data collections
                        return False
        elif right in CaseRightSet.CONTENT.value:
            # Check if the case or case set can be read or written from any of the current data collections
            if is_create_or_delete:
                raise exc.InvalidArgumentsError(
                    f"is_create_or_delete must be False for right {right.value}"
                )
            if tgt_data_collection_ids:
                raise exc.InvalidArgumentsError(
                    f"tgt_data_collection_ids must be empty for right {right.value}"
                )
            has_right_fn = self._get_has_right_function(right)
            for data_collection_id in current_data_collection_ids:
                if data_collection_id in access_abac and has_right_fn(
                    access_abac[data_collection_id]
                ):
                    # Access right in this data collection
                    return True
        else:
            raise exc.InvalidArgumentsError(f"Right {right.value} is invalid")
        # All checks passed
        return True

    def get_case_rights(
        self,
        case_id: UUID,
        case_type_id: UUID,
        created_in_data_collection_id: UUID,
        data_collection_ids: frozenset[UUID],
    ) -> CaseRights:
        case_rights: CaseRights = (
            self._get_case_or_set_rights(  # type:ignore[assignment]
                case_id,
                False,
                case_type_id,
                created_in_data_collection_id,
                data_collection_ids,
            )
        )
        return case_rights

    def get_case_set_rights(
        self,
        case_set_id: UUID,
        case_type_id: UUID,
        created_in_data_collection_id: UUID,
        data_collection_ids: frozenset[UUID],
    ) -> CaseSetRights:
        case_set_rights: CaseSetRights = (
            self._get_case_or_set_rights(  # type:ignore[assignment]
                case_set_id,
                True,
                case_type_id,
                created_in_data_collection_id,
                data_collection_ids,
            )
        )
        return case_set_rights

    def _get_case_or_set_rights(
        self,
        case_or_set_id: UUID,
        is_case_set: bool,
        case_type_id: UUID,
        created_in_data_collection_id: UUID,
        data_collection_ids: frozenset[UUID],
    ) -> CaseRights | CaseSetRights:
        # Parse input
        shared_in_data_collection_ids = data_collection_ids - {
            created_in_data_collection_id
        }
        if self.is_full_access:
            if is_case_set:
                return CaseSetRights(
                    case_set_id=case_or_set_id,
                    case_type_id=case_type_id,
                    created_in_data_collection_id=created_in_data_collection_id,
                    data_collection_ids=data_collection_ids,  # type:ignore[arg-type]
                    is_full_access=True,
                    add_data_collection_ids=set(),
                    remove_data_collection_ids=set(),
                    read_case_set=True,
                    write_case_set=True,
                    can_delete=True,
                    shared_in_data_collection_ids=shared_in_data_collection_ids,
                )
            return CaseRights(
                case_id=case_or_set_id,
                case_type_id=case_type_id,
                created_in_data_collection_id=created_in_data_collection_id,
                data_collection_ids=data_collection_ids,  # type:ignore[arg-type]
                is_full_access=True,
                add_data_collection_ids=set(),
                remove_data_collection_ids=set(),
                read_case_type_col_ids=set(),
                write_case_type_col_ids=set(),
                can_delete=True,
                shared_in_data_collection_ids=shared_in_data_collection_ids,
            )

        # Determine case access: if the case/set created_in_data_collection_id is a
        # private data collection, the user is allowed add to/remove from the
        # listed data collections
        access: dict[UUID, CaseTypeAccessAbac] = self.case_type_access_abacs.get(
            case_type_id, {}
        )
        is_own_private = any(
            x.data_collection_id == created_in_data_collection_id and x.is_private
            for x in self.case_type_access_abacs.get(case_type_id, {}).values()
        )
        # Data collections that the case/set is not yet in but is allowed to be added to
        add_data_collection_ids = (
            {
                x
                for x, y in access.items()
                if (y.add_case_set if is_case_set else y.add_case)
                and x not in data_collection_ids
            }
            if is_own_private
            else set()
        )
        # Data collections that the case/set is in and is allowed to be removed from
        remove_data_collection_ids = (
            {
                x
                for x, y in access.items()
                if (y.remove_case_set if is_case_set else y.remove_case)
                and x in data_collection_ids
            }
            if is_own_private
            else set()
        )

        # Read/write rights
        if is_case_set:
            # Whether the case set can be read/written
            read_case_set = any(x.read_case_set for x in access.values())
            write_case_set = any(x.write_case_set for x in access.values())
        else:
            # Case type cols that can be read/written
            read_case_type_col_ids = set.union(
                *[x.read_case_type_col_ids for x in access.values()]
            )
            write_case_type_col_ids = set.union(
                *[x.write_case_type_col_ids for x in access.values()]
            )

        # Determine any other share rights
        share: dict[UUID, CaseTypeShareAbac] = self.case_type_share_abacs.get(
            case_type_id, {}
        )
        for to_data_collection_id, case_type_share_abac in share.items():
            add_from_data_collection_ids = (
                case_type_share_abac.add_case_set_from_data_collection_ids
                if is_case_set
                else case_type_share_abac.add_case_from_data_collection_ids
            )
            if (
                to_data_collection_id not in data_collection_ids
                and add_from_data_collection_ids.intersection(data_collection_ids)
            ):
                add_data_collection_ids.add(to_data_collection_id)
            remove_from_data_collection_ids = (
                case_type_share_abac.remove_case_set_from_data_collection_ids
                if is_case_set
                else case_type_share_abac.remove_case_from_data_collection_ids
            )
            if (
                to_data_collection_id in data_collection_ids
                and remove_from_data_collection_ids.intersection(data_collection_ids)
            ):
                remove_data_collection_ids.add(to_data_collection_id)

        can_delete = self.is_full_access or set(data_collection_ids).issubset(
            set(remove_data_collection_ids)
        )

        # Create the rights object
        if is_case_set:
            return CaseSetRights(
                case_set_id=case_or_set_id,
                case_type_id=case_type_id,
                created_in_data_collection_id=created_in_data_collection_id,
                data_collection_ids=data_collection_ids,  # type:ignore[arg-type]
                is_full_access=self.is_full_access,
                add_data_collection_ids=add_data_collection_ids,
                remove_data_collection_ids=remove_data_collection_ids,
                read_case_set=read_case_set,
                write_case_set=write_case_set,
                can_delete=can_delete,
                shared_in_data_collection_ids=shared_in_data_collection_ids,
            )
        return CaseRights(
            case_id=case_or_set_id,
            case_type_id=case_type_id,
            created_in_data_collection_id=created_in_data_collection_id,
            data_collection_ids=data_collection_ids,  # type:ignore[arg-type]
            is_full_access=self.is_full_access,
            add_data_collection_ids=add_data_collection_ids,
            remove_data_collection_ids=remove_data_collection_ids,
            read_case_type_col_ids=read_case_type_col_ids,
            write_case_type_col_ids=write_case_type_col_ids,
            can_delete=can_delete,
            shared_in_data_collection_ids=shared_in_data_collection_ids,
        )

    @staticmethod
    def _get_has_right_function(
        right: CaseRight,
    ) -> Callable[[CaseTypeAccessAbac | CaseTypeShareAbac], bool]:
        if right == CaseRight.ADD_CASE:
            has_right_fn = lambda x: x.add_case
        elif right == CaseRight.REMOVE_CASE:
            has_right_fn = lambda x: x.remove_case
        elif right == CaseRight.READ_CASE:
            has_right_fn = lambda x: len(x.read_case_type_col_ids) > 0
        elif right == CaseRight.WRITE_CASE:
            has_right_fn = lambda x: len(x.write_case_type_col_ids) > 0
        elif right == CaseRight.ADD_CASE_SET:
            has_right_fn = lambda x: x.add_case_set
        elif right == CaseRight.REMOVE_CASE_SET:
            has_right_fn = lambda x: x.remove_case_set
        elif right == CaseRight.READ_CASE_SET:
            has_right_fn = lambda x: x.read_case_set
        elif right == CaseRight.WRITE_CASE_SET:
            has_right_fn = lambda x: x.write_case_set
        else:
            raise NotImplementedError(f"Right {right.value} not implemented")
        return has_right_fn

    @staticmethod
    def _get_get_share_from_data_collections_function(
        right: CaseRight,
    ) -> Callable[[CaseTypeShareAbac], set[UUID]]:
        if right == CaseRight.ADD_CASE:
            get_share_from_data_collections_fn = (
                lambda x: x.add_case_from_data_collection_ids
            )
        elif right == CaseRight.REMOVE_CASE:
            get_share_from_data_collections_fn = (
                lambda x: x.remove_case_from_data_collection_ids
            )
        elif right == CaseRight.ADD_CASE_SET:
            get_share_from_data_collections_fn = (
                lambda x: x.add_case_set_from_data_collection_ids
            )
        elif right == CaseRight.REMOVE_CASE_SET:
            get_share_from_data_collections_fn = (
                lambda x: x.remove_case_set_from_data_collection_ids
            )
        else:
            raise NotImplementedError(f"Right {right.value} not implemented")
        return get_share_from_data_collections_fn

    @staticmethod
    def _get_from_data_collections_for_right_function(
        right: CaseRight,
    ) -> Callable[[CaseTypeShareAbac], set[UUID]]:
        if right == CaseRight.ADD_CASE:
            get_from_data_collections_fn = lambda x: x.add_case_from_data_collection_ids
        elif right == CaseRight.REMOVE_CASE:
            get_from_data_collections_fn = (
                lambda x: x.remove_case_from_data_collection_ids
            )
        elif right == CaseRight.ADD_CASE_SET:
            get_from_data_collections_fn = (
                lambda x: x.add_case_set_from_data_collection_ids
            )
        elif right == CaseRight.REMOVE_CASE_SET:
            get_from_data_collections_fn = (
                lambda x: x.remove_case_set_from_data_collection_ids
            )
        else:
            raise NotImplementedError(f"Right {right.value} not implemented")
        return get_from_data_collections_fn



from collections.abc import Callable
from typing import Any
from uuid import UUID

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, enum, model
from gen_epix.commondb.domain.policy import BaseIsOrganizationAdminPolicy
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.fastapp import CrudOperation, CrudOperationSet, exc


class IsOrganizationAdminPolicy(BaseIsOrganizationAdminPolicy):

    def __init__(self, abac_service: BaseAbacService, **kwargs: Any):
        super().__init__(abac_service, **kwargs)

        app_impl: AppImplDetails = abac_service.app.impl
        self.user_class: type[model.User] = app_impl.get_mapped_class(model.User)
        self.role_map = app_impl.role_map
        self.role_set_map = app_impl.role_set_map

        self._get_organization_ids_handler_map: dict[
            type[command.Command],
            Callable[[command.Command], set[UUID]],
        ] = {}
        f = self.register_retrieve_organization_ids_handler
        f(command.SiteCrudCommand, self._get_organization_ids_for_site)
        f(command.ContactCrudCommand, self._get_organization_ids_for_contact)

    def is_allowed(self, cmd: command.Command) -> bool:  # type: ignore[override]
        if cmd.user is None:
            return False
        user: model.User = cmd.user

        # Role is org admin without further ABAC restrictions
        if user.roles.intersection(self.role_set_map[enum.RoleSet.GE_ORG_ADMIN]):
            return True

        # Policy only applies to write operations for crud commands
        if (
            isinstance(cmd, command.CrudCommand)
            and cmd.operation not in CrudOperationSet.WRITE.value
        ):
            return True

        organization_ids = self.retrieve_organization_ids(cmd)
        # Check if user is an admin for all of the affected organizations
        user_admin_organization_ids = (
            self.abac_service.retrieve_organizations_under_admin(
                command.RetrieveOrganizationsUnderAdminCommand(user=user)
            )
        )
        has_permission = organization_ids.issubset(user_admin_organization_ids)
        return has_permission

    def register_retrieve_organization_ids_handler(
        self,
        command_class: type[command.Command],
        handler: Callable[[command.Command], set[UUID]],
    ) -> None:
        self._get_organization_ids_handler_map[command_class] = handler

    def retrieve_organization_ids(self, cmd: command.Command) -> set[UUID]:
        command_class = type(cmd)
        handler = self._get_organization_ids_handler_map.get(command_class)
        if handler is None:
            # Check if handler registered for parent class
            for (
                handler_command_class,
                handler,
            ) in self._get_organization_ids_handler_map.items():
                if issubclass(command_class, handler_command_class):
                    # Parent class handler found -> register it for the child class as well for next time
                    self._get_organization_ids_handler_map[command_class] = handler
                    return handler(cmd)
            raise exc.InitializationServiceError(
                f"No handler registered for command: {command_class}"
            )
        return handler(cmd)

    def _get_organization_ids_for_site(self, cmd: command.SiteCrudCommand) -> set[UUID]:
        sites: list[model.Site] = cmd.get_objs()  # type: ignore[assignment]
        return {x.organization_id for x in sites}

    def _get_organization_ids_for_contact(
        self, cmd: command.ContactCrudCommand
    ) -> set[UUID]:
        contacts: list[model.Contact] = cmd.get_objs()  # type: ignore[assignment]
        sites: list[model.Site] = self.abac_service.app.handle(
            command.ContactCrudCommand(
                user=cmd.user,
                objs=None,
                obj_ids=list(set(x.site_id for x in contacts if x.site_id)),
                operation=CrudOperation.READ_SOME,
            )
        )
        return {x.organization_id for x in sites}



from gen_epix.commondb.domain.command.base import Command

# Non-CRUD commands


class RetrieveOwnPermissionsCommand(Command):
    pass


class RetrieveSubRolesCommand(Command):
    """
    Retrieves all sub-roles of a user. If the current user roles themselves have some
    sub-roles of each other, those will be included as well.
    """

    pass


# CRUD commands



import abc

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.casedb.domain.repository import BaseGeoRepository
from gen_epix.fastapp import BaseService


class BaseGeoService(BaseService):
    SERVICE_TYPE = ServiceType.GEO

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseGeoRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        self.register_default_crud_handlers()
        f(command.RetrieveContainingRegionCommand, self.retrieve_containing_region)

    @abc.abstractmethod
    def retrieve_containing_region(
        self, cmd: command.RetrieveContainingRegionCommand
    ) -> list[model.Region | None]:
        raise NotImplementedError()



# pylint: disable=protected-access
# because the functions are dynamically generated in _is_valid

from __future__ import annotations

from collections.abc import Callable, Generator, Hashable, Iterable, Iterator
from typing import Any, Literal, Self

from pydantic import BaseModel, Field, model_validator

from gen_epix.filter import enum
from gen_epix.filter.base import Filter
from gen_epix.filter.date_range import DateRangeFilter, TypedDateRangeFilter
from gen_epix.filter.datetime_range import DatetimeRangeFilter, TypedDatetimeRangeFilter
from gen_epix.filter.equals_boolean import EqualsBooleanFilter, TypedEqualsBooleanFilter
from gen_epix.filter.equals_number import EqualsNumberFilter, TypedEqualsNumberFilter
from gen_epix.filter.equals_string import EqualsStringFilter, TypedEqualsStringFilter
from gen_epix.filter.equals_uuid import EqualsUuidFilter, TypedEqualsUuidFilter
from gen_epix.filter.exists import ExistsFilter, TypedExistsFilter
from gen_epix.filter.no_filter import NoFilter, TypedNoFilter
from gen_epix.filter.number_range import NumberRangeFilter, TypedNumberRangeFilter
from gen_epix.filter.number_set import NumberSetFilter, TypedNumberSetFilter
from gen_epix.filter.partial_date_range import (
    PartialDateRangeFilter,
    TypedPartialDateRangeFilter,
)
from gen_epix.filter.regex import RegexFilter, TypedRegexFilter
from gen_epix.filter.string_set import StringSetFilter, TypedStringSetFilter
from gen_epix.filter.uuid_set import TypedUuidSetFilter, UuidSetFilter


class CompositeFilter(Filter):
    filters: list[
        ExistsFilter
        | EqualsBooleanFilter
        | EqualsNumberFilter
        | EqualsStringFilter
        | EqualsUuidFilter
        | NumberRangeFilter
        | DateRangeFilter
        | DatetimeRangeFilter
        | PartialDateRangeFilter
        | RegexFilter
        | NumberSetFilter
        | StringSetFilter
        | UuidSetFilter
        | NoFilter
        | CompositeFilter
    ] = Field(description="The list of filters.", min_length=1, frozen=True)
    key: str | None = Field(default=None)
    operator: enum.LogicalOperator = Field(
        default=enum.LogicalOperator.AND,
        description="The boolean operator for the composite filter.",
        frozen=True,
    )
    _is_composite: bool = True

    def _get_row_value(
        self, row: dict | BaseModel, key: Hashable, is_model: bool
    ) -> Any:
        """Get the value from the row, handling both dict and BaseModel."""
        if is_model:
            return getattr(row, key, None)
        return row.get(key, None)

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if len(self.filters) == 0:
            raise AssertionError("At least one filter must be set.")
        if self.operator == enum.LogicalOperator.NOT and len(self.filters) != 1:
            raise AssertionError("Only one filter may be set for NOT operator.")
        if len(self.filters) > 2 and self.operator not in {
            enum.LogicalOperator.AND,
            enum.LogicalOperator.OR,
        }:
            raise AssertionError("operator must be AND or OR for more than 2 filters.")
        # Generate the function to check if a value matches the composite filter
        # The function is generated instead of defined to be able to optimize the check
        # Analogously, generate the function to check if a separate value for each
        # filter
        if self.operator == enum.LogicalOperator.NOT:
            self._match = lambda x: not self.filters[0]._match(x)  # type: ignore
            self._match_row = lambda x, y: x and not self.filters[0]._match(next(y))  # type: ignore
        elif self.operator == enum.LogicalOperator.AND:
            self._match = lambda x: all(filter._match(x) for filter in self.filters)  # type: ignore
            # TODO: improve performance by not using filter.match_row for nested composite filter
            self._match_row = lambda x, y: all(  # type: ignore
                a
                and (filter.match_row(b) if filter._is_composite else filter._match(b))
                for a, b, filter in zip(x, y, self.filters)
            )
        elif self.operator == enum.LogicalOperator.OR:
            self._match = lambda x: any(filter._match(x) for filter in self.filters)  # type: ignore
            # TODO: improve performance by not using filter.match_row for nested composite filter
            self._match_row = lambda x, y: any(  # type: ignore
                a
                and (filter.match_row(b) if filter._is_composite else filter._match(b))
                for a, b, filter in zip(x, y, self.filters)
            )
        elif self.operator == enum.LogicalOperator.XOR:
            self._match = lambda x: self.filters[0]._match(x) != self.filters[1]._match(  # type: ignore
                x
            )
            self._match_row = lambda x, y: all(x) and self.filters[0]._match(  # type: ignore
                next(y)  # type: ignore
            ) != self.filters[
                1
            ]._match(
                next(y)  # type: ignore
            )
        elif self.operator == enum.LogicalOperator.NAND:
            self._match = lambda x: not (  # type: ignore
                self.filters[0]._match(x) and self.filters[1]._match(x)
            )
            self._match_row = lambda x, y: all(x) and not (  # type: ignore
                self.filters[0]._match(next(y)) and self.filters[1]._match(next(y))  # type: ignore
            )
        elif self.operator == enum.LogicalOperator.NOR:
            self._match = lambda x: not (  # type: ignore
                self.filters[0]._match(x) or self.filters[1]._match(x)
            )
            self._match_row = lambda x, y: all(x) and not (  # type: ignore
                self.filters[0]._match(next(y)) or self.filters[1]._match(next(y))  # type: ignore
            )
        elif self.operator == enum.LogicalOperator.XNOR:
            self._match = lambda x: self.filters[0]._match(x) == self.filters[1]._match(  # type: ignore
                x
            )
            self._match_row = lambda x, y: all(x) and self.filters[0]._match(  # type: ignore
                next(y)  # type: ignore
            ) == self.filters[
                1
            ]._match(
                next(y)  # type: ignore
            )
        elif self.operator == enum.LogicalOperator.IMPLIES:
            self._match = lambda x: (  # type: ignore
                not self.filters[0]._match(x) or self.filters[1]._match(x)
            )
            self._match_row = lambda x, y: all(x) and (  # type: ignore
                not self.filters[0]._match(next(y)) or self.filters[1]._match(next(y))  # type: ignore
            )
        elif self.operator == enum.LogicalOperator.NIMPLIES:
            self._match = lambda x: (  # type: ignore
                self.filters[0]._match(x) and not self.filters[1]._match(x)
            )
            self._match_row = lambda x, y: all(x) and (  # type: ignore
                self.filters[0]._match(next(y)) and not self.filters[1]._match(next(y))  # type: ignore
            )

        return self

    def _match(self, value: Any) -> bool:
        # Function is implemented dynamically in _validate_state
        raise NotImplementedError()

    def _match_row(
        self, value_exists: Iterable[bool], value: Iterable[Any], is_model: bool
    ) -> bool:
        # Function is implemented dynamically in _validate_state
        raise NotImplementedError()

    def _not_none_row_iterator(
        self, row: dict[Hashable, Any | None] | BaseModel, is_model: bool = False
    ) -> Generator:
        for filter in self.filters:  # type: ignore
            if filter._is_composite:
                yield all(filter._not_none_row_iterator(row, is_model))
            else:
                yield (
                    (is_model or filter.key in row)
                    and self._get_row_value(row, filter.key, is_model) is not None
                )

    def _not_na_row_iterator(
        self,
        row: dict[Hashable, Any | None] | BaseModel,
        na_values: set[Any],
        is_model: bool = False,
    ) -> Generator:
        for filter in self.filters:  # type: ignore
            if filter._is_composite:
                yield all(filter._not_na_row_iterator(row, na_values, is_model))
            else:
                yield (
                    (is_model or filter.key in row)
                    and self._get_row_value(row, filter.key, is_model) not in na_values
                )

    def _all_subfilters_have_key(self) -> bool:
        retval = True
        for filter in self.filters:
            if filter._is_composite:
                retval = retval and filter._all_subfilters_have_key()
            else:
                retval = retval and filter.key is not None
            if not retval:
                return retval
        return retval

    def _get_map_fun_list(
        self,
        map_fn: (
            dict[Hashable, Callable[[Any], Any]]
            | Callable[[Any], Any]
            | list[Callable[[Any], Any]]
            | None
        ) = None,
    ) -> list[Callable[[Any], Any]]:
        if not map_fn:
            map_fn = [lambda x: x for _ in self.filters]
        elif isinstance(map_fn, dict):
            map_fn = [map_fn.get(x.key, lambda x: x) for x in self.filters]
        elif not isinstance(map_fn, list):
            map_fn = [map_fn for _ in self.filters]
        if not isinstance(map_fn, list):
            raise ValueError("map_fn must be a callable, a list or a dict.")
        if len(map_fn) != len(self.filters):
            raise ValueError("map_fn must have the same length as the filters list.")
        return map_fn

    def match_row(
        self,
        row: dict[Hashable, Any | None] | BaseModel,
        na_values: set[Any] | None = None,
        map_fn: (
            Callable[[Any], Any]
            | dict[Hashable, Callable[[Any], Any]]
            | list[Callable[[Any], Any]]
            | None
        ) = None,
        is_model: bool = False,
    ) -> bool:
        if not self._all_subfilters_have_key():
            raise ValueError(
                "Key must be set for each filter to apply filter to a row."
            )
        # Match, per filter, if both key exists, value not null and value matches
        map_fn = self._get_map_fun_list(map_fn)
        if na_values is None:
            return (
                self._match_row(
                    self._not_none_row_iterator(row, is_model),
                    (
                        (
                            y(row)
                            if x._is_composite
                            else y(self._get_row_value(row, x.key, is_model))
                        )
                        for x, y in zip(self.filters, map_fn)
                    ),
                )
                ^ self.invert
            )
            # yield (
            #     key in row_dict
            #     and row_dict[key] is not None
            #     and self._match(map_fn(row_dict[key]))
            # ) ^ self.invert
        else:
            return (
                self._match_row(
                    self._not_na_row_iterator(row, na_values, is_model),
                    (
                        (
                            y(row)
                            if x._is_composite
                            else y(self._get_row_value(row, x.key, is_model))
                        )
                        for x, y in zip(self.filters, map_fn)
                    ),
                )
                ^ self.invert
            )

    def match_rows(
        self,
        rows: Iterable[dict[Hashable, Any | None] | BaseModel],
        na_values: set[Any] | None = None,
        map_fn: (
            dict[Hashable, Callable[[Any], Any]]
            | Callable[[Any], Any]
            | list[Callable[[Any], Any]]
            | None
        ) = None,
        is_model: bool = False,
    ) -> Iterator[bool]:
        # Match, per row and filter, if both key exists, value not null and value matches
        if not self._all_subfilters_have_key():
            raise ValueError(
                "Key must be set for each filter to apply filter to a row."
            )
        map_fn = self._get_map_fun_list(map_fn)
        if na_values is None:
            for row in rows:
                yield (
                    self._match_row(
                        self._not_none_row_iterator(row, is_model),
                        (
                            (
                                y(row)
                                if x._is_composite
                                else y(self._get_row_value(row, x.key, is_model))
                            )
                            for x, y in zip(self.filters, map_fn)
                        ),
                    )
                    ^ self.invert
                )
        else:
            for row in rows:
                yield (
                    self._match_row(
                        self._not_na_row_iterator(row, na_values, is_model),
                        (
                            (
                                y(row)
                                if x._is_composite
                                else y(self._get_row_value(row, x.key, is_model))
                            )
                            for x, y in zip(self.filters, map_fn)
                        ),
                    )
                    ^ self.invert
                )

    def filter_rows(
        self,
        rows: Iterable[dict[Hashable, Any | None] | BaseModel],
        na_values: set[Any] | None = None,
        map_fn: (
            dict[Hashable, Callable[[Any], Any]]
            | Callable[[Any], Any]
            | list[Callable[[Any], Any]]
            | None
        ) = None,
        is_model: bool = False,
    ) -> Iterator[dict[Hashable, Any | None]]:
        # Match, per row and filter, if both key exists, value not null and value matches
        if not self._all_subfilters_have_key():
            raise ValueError(
                "Key must be set for each filter to apply filter to a row."
            )
        map_fn = self._get_map_fun_list(map_fn)
        if na_values is None:
            for row in rows:
                if (
                    self._match_row(
                        self._not_none_row_iterator(row, is_model),
                        (
                            (
                                y(row)
                                if x._is_composite
                                else y(self._get_row_value(row, x.key, is_model))
                            )
                            for x, y in zip(self.filters, map_fn)
                        ),
                    )
                    ^ self.invert
                ):
                    yield row
        else:
            for row in rows:
                if (
                    self._match_row(
                        self._not_na_row_iterator(row, na_values, is_model),
                        (
                            (
                                y(row)
                                if x._is_composite
                                else y(self._get_row_value(row, x.key, is_model))
                            )
                            for x, y in zip(self.filters, map_fn)
                        ),
                    )
                    ^ self.invert
                ):
                    yield row

    def get_keys(self) -> list[Hashable]:
        keys = []

        def _recursion(keys: list, filters: list[Filter]) -> None:
            for filter in filters:
                if isinstance(filter, CompositeFilter):
                    _recursion(keys, filter.filters)
                else:
                    keys.append(filter.key)

        _recursion(keys, self.filters)
        return keys

    def set_keys(
        self, key_map: dict[Hashable, Hashable] | Callable[[Hashable], Hashable]
    ) -> Self:
        for filter in self.filters:
            if isinstance(filter, CompositeFilter):
                filter.set_keys(key_map)
            elif isinstance(key_map, dict):
                filter.set_key(key_map.get(filter.key, filter.key))
            else:
                filter.set_key(key_map)
        return self


class TypedCompositeFilter(CompositeFilter):
    type: Literal[enum.FilterType.COMPOSITE.value]
    filters: list[
        TypedExistsFilter
        | TypedEqualsBooleanFilter
        | TypedEqualsNumberFilter
        | TypedEqualsStringFilter
        | TypedEqualsUuidFilter
        | TypedNumberRangeFilter
        | TypedDateRangeFilter
        | TypedDatetimeRangeFilter
        | TypedPartialDateRangeFilter
        | TypedRegexFilter
        | TypedNumberSetFilter
        | TypedStringSetFilter
        | TypedUuidSetFilter
        | TypedNoFilter
        | TypedCompositeFilter,
    ] = Field(
        description="The list of filters."
    )  # type: ignore



import gen_epix.casedb.domain.command as command
import gen_epix.casedb.domain.enum as enum
import gen_epix.casedb.domain.model as model
from gen_epix.casedb.domain import exc
from gen_epix.casedb.domain.policy import BaseCaseAbacPolicy
from gen_epix.casedb.domain.service import BaseCaseService as DomainBaseCaseService
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.fastapp import CrudOperation


def case_service_create_cases(
    self: BaseCaseService, cmd: command.CreateCasesCommand
) -> list[model.Case] | None:
    # Special case: zero cases to be created
    if not cmd.cases:
        return []

    # Get case type and created_in data collection IDs
    case_type_id = cmd.case_type_id
    created_in_data_collection_id = cmd.created_in_data_collection_id

    # @ABAC: verify if case set or cases may be created in the given data collection(s)
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None
    is_allowed = case_abac.is_allowed(
        case_type_id,
        enum.CaseRight.ADD_CASE,
        True,
        created_in_data_collection_id=created_in_data_collection_id,
        tgt_data_collection_ids=cmd.data_collection_ids,
    )
    if not is_allowed:
        assert cmd.user is not None
        raise exc.UnauthorizedAuthError(
            f"User {cmd.user.id} is not allowed to create cases in the given data collection(s)"
        )

    # Convert cases for create update to cases
    # TODO: validate content and add derived values
    cases: list[model.Case] = [
        model.Case(
            id=x.id,
            case_type_id=cmd.case_type_id,
            subject_id=x.subject_id,
            created_in_data_collection_id=cmd.created_in_data_collection_id,
            case_date=x.case_date,
            content={y: z for y, z in x.content.items() if z is not None},
        )
        for x in cmd.cases
    ]

    # Create cases and case data collection links
    with self.repository.uow() as uow:
        # Create cases, using the parent class method to avoid ABAC
        # restrictions
        cases = super(DomainBaseCaseService, self).crud(  # type: ignore[assignment]
            command.CaseCrudCommand(
                user=cmd.user,
                operation=CrudOperation.CREATE_SOME,
                objs=cases,  # type: ignore[arg-type]
                props=cmd.props,
            )
        )
        # Associate cases with data collections
        curr_cmd = command.CaseDataCollectionLinkCrudCommand(
            user=cmd.user,
            operation=CrudOperation.CREATE_SOME,
            objs=[
                model.CaseDataCollectionLink(
                    case_id=x.id, data_collection_id=y  # type: ignore[arg-type]
                )
                for x in cases
                for y in cmd.data_collection_ids
            ],
        )
        curr_cmd._policies.extend(cmd._policies)
        case_data_collection_links = self.crud(curr_cmd)
    return cases


def case_service_create_case_set(
    self: BaseCaseService, cmd: command.CreateCaseSetCommand
) -> model.CaseSet | None:
    # Get case type and created_in data collection IDs
    case_type_id = cmd.case_set.case_type_id
    created_in_data_collection_id = cmd.case_set.created_in_data_collection_id

    # @ABAC: verify if case set or cases may be created in the given data collection(s)
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None
    is_allowed = case_abac.is_allowed(
        case_type_id,
        enum.CaseRight.ADD_CASE_SET,
        True,
        created_in_data_collection_id=created_in_data_collection_id,
        tgt_data_collection_ids=cmd.data_collection_ids,
    )
    if not is_allowed:
        assert cmd.user is not None
        raise exc.UnauthorizedAuthError(
            f"User {cmd.user.id} is not allowed to create a case set in the given data collection(s)"
        )

    # Create case set, case set data collection links, and optionally
    # case set members
    with self.repository.uow() as uow:
        # Create case set, using the parent class method to avoid ABAC
        # restrictions
        case_set: model.CaseSet = super(DomainBaseCaseService, self).crud(  # type: ignore[assignment]
            command.CaseSetCrudCommand(
                user=cmd.user,
                operation=CrudOperation.CREATE_ONE,
                objs=cmd.case_set,
                props=cmd.props,
            )
        )
        # Associate case set/cases with data collections
        assert case_set.id is not None
        curr_cmd = command.CaseSetDataCollectionLinkCrudCommand(
            user=cmd.user,
            operation=CrudOperation.CREATE_SOME,
            objs=[
                model.CaseSetDataCollectionLink(
                    case_set_id=case_set.id, data_collection_id=x
                )
                for x in cmd.data_collection_ids
            ],
        )
        curr_cmd._policies.extend(cmd._policies)
        case_set_data_collection_links = self.crud(curr_cmd)
        # Associate case set with cases if necessary
        if cmd.case_ids:
            curr_cmd2: command.CaseSetMemberCrudCommand = (
                command.CaseSetMemberCrudCommand(
                    user=cmd.user,
                    operation=CrudOperation.CREATE_SOME,
                    objs=[
                        model.CaseSetMember(case_set_id=case_set.id, case_id=x)
                        for x in cmd.case_ids
                    ],
                )
            )
            curr_cmd2._policies.extend(cmd._policies)
            case_set_members = self.crud(curr_cmd2)
    return case_set



# pylint: disable=useless-import-alias
from gen_epix import fastapp
from gen_epix.casedb.domain import enum
from gen_epix.casedb.domain.model.abac import CaseAbac as CaseAbac
from gen_epix.casedb.domain.model.abac import CaseTypeAccessAbac as CaseTypeAccessAbac
from gen_epix.casedb.domain.model.abac import CaseTypeShareAbac as CaseTypeShareAbac
from gen_epix.casedb.domain.model.abac import (
    OrganizationAccessCasePolicy as OrganizationAccessCasePolicy,
)
from gen_epix.casedb.domain.model.abac import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)
from gen_epix.casedb.domain.model.abac import (
    OrganizationShareCasePolicy as OrganizationShareCasePolicy,
)
from gen_epix.casedb.domain.model.abac import (
    UserAccessCasePolicy as UserAccessCasePolicy,
)
from gen_epix.casedb.domain.model.abac import UserShareCasePolicy as UserShareCasePolicy
from gen_epix.casedb.domain.model.case import BaseCaseRights as BaseCaseRights
from gen_epix.casedb.domain.model.case import Case as Case
from gen_epix.casedb.domain.model.case import (
    CaseDataCollectionLink as CaseDataCollectionLink,
)
from gen_epix.casedb.domain.model.case import CaseDataIssue as CaseDataIssue
from gen_epix.casedb.domain.model.case import CaseQuery as CaseQuery
from gen_epix.casedb.domain.model.case import CaseReadSet as CaseReadSet
from gen_epix.casedb.domain.model.case import CaseRights as CaseRights
from gen_epix.casedb.domain.model.case import CaseSeq as CaseSeq
from gen_epix.casedb.domain.model.case import CaseSet as CaseSet
from gen_epix.casedb.domain.model.case import CaseSetCategory as CaseSetCategory
from gen_epix.casedb.domain.model.case import (
    CaseSetDataCollectionLink as CaseSetDataCollectionLink,
)
from gen_epix.casedb.domain.model.case import CaseSetMember as CaseSetMember
from gen_epix.casedb.domain.model.case import CaseSetQuery as CaseSetQuery
from gen_epix.casedb.domain.model.case import CaseSetRights as CaseSetRights
from gen_epix.casedb.domain.model.case import CaseSetStat as CaseSetStat
from gen_epix.casedb.domain.model.case import CaseSetStatus as CaseSetStatus
from gen_epix.casedb.domain.model.case import CaseType as CaseType
from gen_epix.casedb.domain.model.case import CaseTypeCol as CaseTypeCol
from gen_epix.casedb.domain.model.case import CaseTypeColSet as CaseTypeColSet
from gen_epix.casedb.domain.model.case import (
    CaseTypeColSetMember as CaseTypeColSetMember,
)
from gen_epix.casedb.domain.model.case import CaseTypeDim as CaseTypeDim
from gen_epix.casedb.domain.model.case import CaseTypeSet as CaseTypeSet
from gen_epix.casedb.domain.model.case import CaseTypeSetCategory as CaseTypeSetCategory
from gen_epix.casedb.domain.model.case import CaseTypeSetMember as CaseTypeSetMember
from gen_epix.casedb.domain.model.case import CaseTypeStat as CaseTypeStat
from gen_epix.casedb.domain.model.case import (
    CaseValidationReport as CaseValidationReport,
)
from gen_epix.casedb.domain.model.case import Col as Col
from gen_epix.casedb.domain.model.case import Dim as Dim
from gen_epix.casedb.domain.model.case import (
    GeneticDistanceProtocol as GeneticDistanceProtocol,
)
from gen_epix.casedb.domain.model.case import TreeAlgorithm as TreeAlgorithm
from gen_epix.casedb.domain.model.case import TreeAlgorithmClass as TreeAlgorithmClass
from gen_epix.casedb.domain.model.case import ValidatedCase as ValidatedCase
from gen_epix.casedb.domain.model.case.complete_case_type import (
    CompleteCaseType as CompleteCaseType,
)
from gen_epix.casedb.domain.model.case.non_persistable import (
    CaseForCreateUpdate as CaseForCreateUpdate,
)
from gen_epix.casedb.domain.model.geo import Region as Region
from gen_epix.casedb.domain.model.geo import RegionRelation as RegionRelation
from gen_epix.casedb.domain.model.geo import RegionSet as RegionSet
from gen_epix.casedb.domain.model.geo import RegionSetShape as RegionSetShape
from gen_epix.casedb.domain.model.ontology import Concept as Concept
from gen_epix.casedb.domain.model.ontology import ConceptRelation as ConceptRelation
from gen_epix.casedb.domain.model.ontology import ConceptSet as ConceptSet
from gen_epix.casedb.domain.model.ontology import Disease as Disease
from gen_epix.casedb.domain.model.ontology import EtiologicalAgent as EtiologicalAgent
from gen_epix.casedb.domain.model.ontology import Etiology as Etiology
from gen_epix.casedb.domain.model.seqdb import AlleleProfile as AlleleProfile
from gen_epix.casedb.domain.model.seqdb import AssemblyProtocol as AssemblyProtocol
from gen_epix.casedb.domain.model.seqdb import File as File
from gen_epix.casedb.domain.model.seqdb import GeneticSequence as GeneticSequence
from gen_epix.casedb.domain.model.seqdb import (
    LibraryPrepProtocol as LibraryPrepProtocol,
)
from gen_epix.casedb.domain.model.seqdb import PhylogeneticTree as PhylogeneticTree
from gen_epix.casedb.domain.model.seqdb import RawSeq as RawSeq
from gen_epix.casedb.domain.model.seqdb import ReadSet as ReadSet
from gen_epix.casedb.domain.model.seqdb import Sample as Sample
from gen_epix.casedb.domain.model.seqdb import Seq as Seq
from gen_epix.casedb.domain.model.subject import Subject as Subject
from gen_epix.casedb.domain.model.subject import SubjectIdentifier as SubjectIdentifier
from gen_epix.commondb.domain import enum as common_enum
from gen_epix.commondb.domain.model import (
    SORTED_MODELS_BY_SERVICE_TYPE as _COMMON_SORTED_MODELS_BY_SERVICE_TYPE,
)
from gen_epix.commondb.domain.model import Contact as Contact
from gen_epix.commondb.domain.model import DataCollection as DataCollection
from gen_epix.commondb.domain.model import DataCollectionSet as DataCollectionSet
from gen_epix.commondb.domain.model import (
    DataCollectionSetMember as DataCollectionSetMember,
)
from gen_epix.commondb.domain.model import IdentifierIssuer as IdentifierIssuer
from gen_epix.commondb.domain.model import Model as Model
from gen_epix.commondb.domain.model import Organization as Organization
from gen_epix.commondb.domain.model import OrganizationSet as OrganizationSet
from gen_epix.commondb.domain.model import (
    OrganizationSetMember as OrganizationSetMember,
)
from gen_epix.commondb.domain.model import Outage as Outage
from gen_epix.commondb.domain.model import Site as Site
from gen_epix.commondb.domain.model import User as User
from gen_epix.commondb.domain.model import UserInvitation as UserInvitation
from gen_epix.commondb.domain.model import (
    UserInvitationConstraints as UserInvitationConstraints,
)
from gen_epix.commondb.domain.model import UserNameEmail as UserNameEmail
from gen_epix.fastapp.services.auth import IdentityProvider as IdentityProvider
from gen_epix.fastapp.services.auth import IDPUser as IDPUser

# List up model classes per service and sorted according to links topology
SORTED_MODELS_BY_SERVICE_TYPE: dict[enum.ServiceType, list[type[fastapp.Model]]] = (
    {  # pyright: ignore[reportAssignmentType]
        # Common models
        enum.ServiceType.AUTH: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.AUTH]
        ),
        enum.ServiceType.SYSTEM: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.SYSTEM]
        ),
        enum.ServiceType.RBAC: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.RBAC]
        ),
        enum.ServiceType.ORGANIZATION: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.ORGANIZATION]
        ),
        # Specific models
        enum.ServiceType.ONTOLOGY: [
            ConceptSet,
            Concept,
            ConceptRelation,
            Disease,
            EtiologicalAgent,
            Etiology,
        ],
        enum.ServiceType.GEO: [
            RegionSet,
            Region,
            RegionRelation,
            RegionSetShape,
        ],
        enum.ServiceType.SEQDB: [
            AlleleProfile,
            GeneticSequence,
            PhylogeneticTree,
            LibraryPrepProtocol,
            ReadSet,
            File,
            RawSeq,
            Sample,
            AssemblyProtocol,
            Seq,
        ],
        enum.ServiceType.SUBJECT: [
            Subject,
            SubjectIdentifier,
        ],
        enum.ServiceType.CASE: [
            TreeAlgorithmClass,
            TreeAlgorithm,
            GeneticDistanceProtocol,
            Dim,
            Col,
            CaseReadSet,
            CaseSeq,
            CaseTypeSetCategory,
            CaseType,
            CaseTypeSet,
            CaseTypeSetMember,
            CaseTypeDim,
            CaseTypeCol,
            CaseTypeColSet,
            CaseTypeColSetMember,
            CompleteCaseType,
            Case,
            CaseForCreateUpdate,
            CaseSetCategory,
            CaseSetStatus,
            CaseSet,
            CaseSetMember,
            CaseDataCollectionLink,
            CaseSetDataCollectionLink,
            CaseTypeStat,
            CaseSetStat,
            CaseQuery,
            CaseSetQuery,
            CaseRights,
            CaseSetRights,
            CaseValidationReport,
        ],
        enum.ServiceType.ABAC: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.ABAC]
        )
        + [
            OrganizationAccessCasePolicy,
            OrganizationShareCasePolicy,
            UserAccessCasePolicy,
            UserShareCasePolicy,
        ],
    }
)
SORTED_SERVICE_TYPES = tuple(SORTED_MODELS_BY_SERVICE_TYPE.keys())

COMMON_MODEL_MAP: dict[type[fastapp.Model], type[fastapp.Model]] = {}



from gen_epix.casedb.domain import model
from gen_epix.casedb.domain.policy import BaseCaseAbacPolicy
from gen_epix.fastapp import Command


class CaseAbacPolicy(BaseCaseAbacPolicy):
    def get_content(self, cmd: Command) -> model.CaseAbac:
        return self.abac_service.get_case_abac(cmd)

    def get_content_return_type(self, cmd: Command) -> type[model.Model]:
        return model.CaseAbac



# pylint: disable=useless-import-alias
from gen_epix.commondb.config.cfg import AppCfg as AppCfg
from gen_epix.commondb.config.cfg import BaseAppCfg as BaseAppCfg
from gen_epix.commondb.config.cfg import SettingsManager as SettingsManager
from gen_epix.commondb.config.factory import IdFactory as IdFactory
from gen_epix.commondb.config.factory import TimestampFactory as TimestampFactory



from gen_epix.casedb.domain.repository import BaseOntologyRepository
from gen_epix.fastapp.repositories import DictRepository


class OntologyDictRepository(DictRepository, BaseOntologyRepository):
    pass



import logging
import ssl
from typing import Annotated, Any

from fastapi import Depends, Request, Security
from fastapi.security import SecurityScopes

from gen_epix.fastapp import App, enum, exc, model
from gen_epix.fastapp.services.auth.base import BaseAuthService
from gen_epix.fastapp.services.auth.command import GetIdentityProvidersCommand
from gen_epix.fastapp.services.auth.idp_client import IdpClient
from gen_epix.fastapp.services.auth.mock_idp_client import MockIDPClient
from gen_epix.fastapp.services.auth.model import (
    Claims,
    IdentityProvider,
    IDPUser,
    OidcServerCfg,
)
from gen_epix.fastapp.services.auth.oauth_idp_client import OauthIdpClient


class AuthService(BaseAuthService):

    DEFAULT_IS_PUBLIC_IDP = False  # Security: IDPs are not public by default

    def __init__(
        self,
        app: App,
        logger: logging.Logger | None = None,
        setup_logger: logging.Logger | None = None,
        idps_cfg: list[dict[str, str | list]] | None = None,
        ssl_context: ssl.SSLContext | bool = True,
        repository: None = None,
        **kwargs: Any,
    ):
        super().__init__(
            app,
            repository=repository,
            logger=logger,
            setup_logger=setup_logger,
            **kwargs,
        )

        # Initialize authentication services
        self._init_idp_clients(app, idps_cfg, ssl_context)
        self._idp_client_by_id = {x.id: x for x in self._idp_clients or []}

        # Initialize no authentication user
        self._no_auth_user: model.User
        self._no_auth_idp_client: IdpClient = MockIDPClient(logger=logger)

    @property
    def idp_clients(self) -> list[IdpClient]:
        return list(self._idp_clients)

    async def get_existing_user_from_token(self, token: str) -> model.User | None:
        for idp_client in self._idp_clients:
            jwt_claims = await idp_client.get_claims_from_jwt(token)
            if jwt_claims:
                try:
                    user = await self.get_existing_user_from_claims(
                        Claims(
                            claims=jwt_claims,
                            scheme="BEARER",
                            token=token,
                            idp_client_id=idp_client.id,
                        )
                    )
                    return user
                except exc.UnauthorizedAuthError:
                    continue
        # No valid user found for any of the IDP clients
        raise exc.UnauthorizedAuthError()

    def create_user_dependencies(
        self,
    ) -> tuple[model.User, model.User, IDPUser]:

        if not self._idp_clients:
            # No authentication -> create/retrieve root user
            user_manager = self.app.user_manager
            if not user_manager:
                raise exc.InitializationServiceError(
                    "No authentication services configured and no user generator provided"
                )
            self._no_auth_user = user_manager.create_root_user_from_claims({})

            async def dummy_get_existing_user(
                request: Request, _security_scopes: SecurityScopes
            ) -> model.User:
                claims = await self._no_auth_idp_client(request)
                if claims:
                    user = await self.get_existing_user_from_claims(
                        claims, request_userinfo=False
                    )
                    if user:
                        return user
                return self._no_auth_user

            async def dummy_get_new_user(
                request: Request, _security_scopes: SecurityScopes
            ) -> model.User:
                claims = await self._no_auth_idp_client(request)
                if claims:
                    user = await self.get_new_user_from_claims(
                        claims, request_userinfo=False
                    )
                    if user:
                        return user
                raise exc.UnauthorizedAuthError(
                    "Unable to create user due to missing header or claims"
                )

            registered_user_dependency: model.User = Annotated[  # type: ignore
                model.User,
                Security(dummy_get_existing_user, scopes=["openid", "profile"]),
            ]
            new_user_dependency: model.User = Annotated[  # type: ignore
                model.User,
                Security(dummy_get_new_user, scopes=["openid", "profile"]),
            ]
            idp_user_dependency: IDPUser = Annotated[  # type: ignore
                IDPUser,
                Security(dummy_get_new_user, scopes=["openid", "profile"]),
            ]

            return registered_user_dependency, new_user_dependency, idp_user_dependency

        # Init get_current_user function definition and environment
        # TODO: generate get_current_user and get_new_user functions
        # dynamically based on number of authentication services
        n_security_bases = len(self._idp_clients)
        if self._logger:
            self._logger.info(
                self.create_log_message(
                    "1e3d75cb",
                    "Generating user authentication function per security base",
                )
            )

        def _warn(request: Request) -> None:
            if self._logger:
                self._logger.warning(
                    self.create_log_message(
                        "f8853e9e",
                        "Unable to verify provided user",
                        request=request,
                    )
                )

        idp_client_list = self._idp_clients + ([None] * (5 - len(self._idp_clients)))

        async def get_current_user1(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user1(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user1(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user2(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user2(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user2(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user3(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            if claims_2:
                return await self.get_existing_user_from_claims(claims_2)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user3(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            if claims_2:
                return await self.get_new_user_from_claims(claims_2)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user3(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            if claims_2:
                return await self.get_idp_user_from_claims(claims_2)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user4(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            if claims_2:
                return await self.get_existing_user_from_claims(claims_2)
            if claims_3:
                return await self.get_existing_user_from_claims(claims_3)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user4(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            if claims_2:
                return await self.get_new_user_from_claims(claims_2)
            if claims_3:
                return await self.get_new_user_from_claims(claims_3)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user4(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            if claims_2:
                return await self.get_idp_user_from_claims(claims_2)
            if claims_3:
                return await self.get_idp_user_from_claims(claims_3)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user5(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
            claims_4: Claims = Depends(idp_client_list[4]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            if claims_2:
                return await self.get_existing_user_from_claims(claims_2)
            if claims_3:
                return await self.get_existing_user_from_claims(claims_3)
            if claims_4:
                return await self.get_existing_user_from_claims(claims_4)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user5(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
            claims_4: Claims = Depends(idp_client_list[4]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            if claims_2:
                return await self.get_new_user_from_claims(claims_2)
            if claims_3:
                return await self.get_new_user_from_claims(claims_3)
            if claims_4:
                return await self.get_new_user_from_claims(claims_4)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user5(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
            claims_4: Claims = Depends(idp_client_list[4]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            if claims_2:
                return await self.get_idp_user_from_claims(claims_2)
            if claims_3:
                return await self.get_idp_user_from_claims(claims_3)
            if claims_4:
                return await self.get_idp_user_from_claims(claims_4)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        get_idp_user_functions = [
            get_idp_user1,
            get_idp_user2,
            get_idp_user3,
            get_idp_user4,
            get_idp_user5,
        ]
        get_current_user_functions = [
            get_current_user1,
            get_current_user2,
            get_current_user3,
            get_current_user4,
            get_current_user5,
        ]
        get_new_user_functions = [
            get_new_user1,
            get_new_user2,
            get_new_user3,
            get_new_user4,
            get_new_user5,
        ]

        # Create CurrentUser/NewUser, injecting get_current_user/get_new_user
        if n_security_bases > len(get_current_user_functions):
            msg = (
                f"More than {len(get_current_user_functions)} "
                "({n_security_bases}) not implemented"
            )
            if self._logger:
                self._logger.error(self.create_log_message("d6f4ede7", msg))
            raise exc.InitializationServiceError(msg)
        registered_user_dependency: model.User = Annotated[  # type: ignore
            model.User,
            Security(
                get_current_user_functions[n_security_bases - 1],  # type: ignore
                scopes=["openid", "profile"],
            ),
        ]
        new_user_dependency: model.User = Annotated[  # type: ignore
            model.User,
            Security(
                get_new_user_functions[n_security_bases - 1],  # type: ignore
                scopes=["openid", "profile"],
            ),
        ]
        idp_user_dependency: IDPUser = Annotated[  # type: ignore
            IDPUser,
            Security(
                get_idp_user_functions[n_security_bases - 1],  # type: ignore
                scopes=["openid", "profile"],
            ),
        ]
        return registered_user_dependency, new_user_dependency, idp_user_dependency

    def get_identity_providers(
        self,
        cmd: GetIdentityProvidersCommand,
    ) -> list[IdentityProvider]:
        identity_providers = [x.get_identity_provider() for x in self._idp_clients]
        if cmd.public:
            return [x for x in identity_providers if x.public]
        return identity_providers

    async def get_idp_user_from_claims(self, claims: Claims) -> IDPUser:
        claims_dict = claims.claims
        issuer: str = claims_dict["iss"]  # type: ignore
        sub: str = claims_dict["sub"]  # type: ignore

        return IDPUser(issuer=issuer, sub=sub)

    async def get_new_user_from_claims(
        self, claims: Claims, request_userinfo: bool = True
    ) -> model.User:
        # Get userinfo
        if request_userinfo:
            claims.claims.update(
                self._idp_client_by_id[claims.idp_client_id].get_claims_from_userinfo(
                    claims.token
                )
            )
        # Create user obj
        user_manager = self.app.user_manager
        if user_manager:
            # Use user manager to create user
            new_user = user_manager.get_user_instance_from_claims(claims.claims)
            if new_user:
                return new_user
        else:
            # No user manager configured, create user obj directly from claims
            new_user = model.User(**claims.claims)  # type: ignore
        return new_user

    async def get_existing_user_from_claims(
        self, claims: Claims, request_userinfo: bool = True
    ) -> model.User:
        issuer: str = claims.claims["iss"]  # type: ignore
        sub: str = claims.claims["sub"]  # type: ignore
        user_manager = self.app.user_manager
        if not user_manager:
            # No user generator configured
            raise exc.UnauthorizedAuthError()

        user_key = user_manager.get_user_key_from_claims(claims.claims)
        if not user_key and request_userinfo:
            claims.claims.update(
                self._idp_client_by_id[claims.idp_client_id].get_claims_from_userinfo(
                    claims.token
                )
            )
            user_key = user_manager.get_user_key_from_claims(claims.claims)
        if not user_key:
            if self._logger:
                self._logger.warning(
                    self.create_log_message(
                        "d3b7e9f1",
                        "No user key found in claims",
                        sub=sub,
                        user_key=user_key,
                    )
                )
            raise exc.UnauthorizedAuthError()

        try:
            # Retrieve existing user
            user = user_manager.retrieve_user_by_key(user_key)
            try:
                # Retrieve user name from claims and update if necessary
                new_user_name = user_manager.get_user_name_from_claims(claims.claims)
                if new_user_name:
                    updated_user = user_manager.update_user_name(user, new_user_name)
                    if updated_user:
                        return updated_user
            except exc.DomainException as exception:
                if self._logger:
                    self._logger.error(
                        self.create_log_message(
                            "a7d93f8c",
                            "Failed to update user name from claims",
                            issuer=issuer,
                            sub=sub,
                            user_key=user_key,
                            exception=exception,
                        )
                    )
            return user

        except exc.NoResultsError:
            # User does not exist
            if self._logger:
                self._logger.warning(
                    self.create_log_message(
                        "ec9625ad",
                        "User not found",
                        issuer=issuer,
                        sub=sub,
                        user_key=user_key,
                    )
                )

            # Check if this is the root user, if so create it and and its dependencies
            if user_manager.is_root_user_claims(claims.claims):
                if self._logger:
                    self._logger.warning(
                        self.create_log_message(
                            "d4f37b85",
                            "User is root user, creating",
                            issuer=issuer,
                            sub=sub,
                            user_key=user_key,
                        )
                    )
                return user_manager.create_root_user_from_claims(claims.claims)

            # Automatically create the user if configured
            try:
                user = user_manager.create_user_from_claims(claims.claims)
                if not user:
                    raise exc.UnauthorizedAuthError()
                if self._logger:
                    self._logger.info(
                        self.create_log_message(
                            "fe8bfbd0",
                            "Automatically created user",
                            issuer=issuer,
                            sub=sub,
                            user_key=user_key,
                        )
                    )
                return user
            except Exception as exception:
                if self._logger:
                    self._logger.error(
                        self.create_log_message(
                            "08e3c18b",
                            "Could not automatically create user",
                            issuer=issuer,
                            sub=sub,
                            user_key=user_key,
                            exception=exception,
                        )
                    )
                raise exc.UnauthorizedAuthError()

    def _init_idp_clients(
        self,
        app: App,
        idps_cfg: list[dict[str, str | list]] | None,
        ssl_context: ssl.SSLContext | bool,
    ) -> None:
        # Initialize list of all IDP clients and list of exposed IDP clients
        if not idps_cfg:
            idps_cfg = []
        idp_clients: list[IdpClient] = []
        exposed_idp_clients: list[IdpClient] = []
        idp_names: set[str] = set()
        idp_labels: set[str] = set()
        logger = app.logger
        for idp_cfg in idps_cfg:
            idp_name: str = idp_cfg["name"]  # type: ignore[assignment]
            idp_label: str = idp_cfg["label"]  # type: ignore[assignment]
            is_public: bool = idp_cfg.get("is_exposed", self.DEFAULT_IS_PUBLIC_IDP)  # type: ignore[assignment]
            if idp_name in idp_names or idp_label in idp_labels:
                msg = (
                    "Authentication service name and/or label are not unique: "
                    f"{idp_cfg['name']}, {idp_cfg['label']}"
                )
                if logger:
                    logger.error(app.create_log_message("30a9a272", msg))
                raise exc.InitializationServiceError(msg)
            idp_names.add(idp_name)
            idp_labels.add(idp_label)

            oidc_discovery_doc_keys = set(OidcServerCfg.model_fields.keys())
            try:
                protocol = enum.AuthProtocol[str(idp_cfg["protocol"])]
                if protocol == enum.AuthProtocol.OIDC:
                    discovery_doc = {
                        x: y for x, y in idp_cfg.items() if x in oidc_discovery_doc_keys
                    }
                    idp_client = OauthIdpClient(
                        OidcServerCfg(**idp_cfg),  # type: ignore
                        logger=self._logger,
                        log_item_class=app.log_item_class,
                        discovery_doc=discovery_doc,  # Provide again to avoid fetching from discovery URL (again).
                        ssl_context=ssl_context,
                    )
                else:
                    raise exc.InitializationServiceError(
                        f"Protocol {protocol.value} not implemented"
                    )
                idp_clients.append(idp_client)
                if is_public:
                    exposed_idp_clients.append(idp_client)
            except Exception as exception:
                # Unable to initialize authentication service: do not raise
                # an error to avoid entire app not starting up
                msg = (
                    "Could not initialize authentication client for service "
                    f"{idp_cfg['name']}"
                )
                if logger:
                    logger.error(
                        app.create_log_message("48b7e021", msg, exception=exception)
                    )
        for idp_client in idp_clients:  # type: ignore
            if isinstance(idp_client, OauthIdpClient):
                if logger:
                    logger.info(
                        app.create_log_message(
                            "7e0b64cc",
                            f"Authentication client on {idp_client.issuer} initialized",
                        )
                    )
            else:
                raise exc.InitializationServiceError(
                    f"Authentication client of type {type(idp_client)} "
                    "not implemented"
                )
        self._idp_clients = idp_clients
        self._exposed_idp_clients = exposed_idp_clients



"""
Conditional transformer implementation.
"""

from collections.abc import Callable

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.transformer import Transformer


class ConditionalTransformer(Transformer):
    """Apply transformation only when condition is met."""

    def __init__(
        self,
        condition: Callable[[ObjectAdapter], bool],
        transformer: Transformer,
        name: str | None = None,
    ):
        super().__init__(name)
        self.condition = condition
        self.transformer = transformer

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Apply transformation if condition is met."""
        if self.condition(obj):
            return self.transformer.transform(obj)
        return obj



from collections.abc import Callable
from functools import partial
from typing import Any

from fastapi.openapi.utils import get_openapi

from gen_epix.fastapp.services.auth import AuthService

_GET_OPEN_API_DEFAULTS: dict[str, Any] = {
    "title": "API",
    "description": "API description",
    "version": "0.0.0",
    "separate_input_output_schemas": False,
}


def create_custom_openapi_function(
    get_open_api_kwargs: dict[str, Any] = {},
    fix_schema: bool = True,
    auth_service: AuthService | None = None,
) -> Callable[[], dict[str, Any]]:

    def custom_openapi_function(
        default_kwargs: dict[str, Any],
        get_open_api_kwargs: dict[str, Any],
        fix_schema: bool,
        auth_service: AuthService | None,
    ) -> dict[str, Any]:
        # Set defaults
        for key, value in _GET_OPEN_API_DEFAULTS.items():
            get_open_api_kwargs[key] = get_open_api_kwargs.get(key, value)

        # Get the initial OpenAPI schema
        openapi_schema = get_openapi(**get_open_api_kwargs)

        # Fix any issues with the schema if necessary
        # TODO: add a fix for read-only fields
        if fix_schema:
            fix_schema_nullable_and_single_element(openapi_schema)

        # Add x-tokenName to security schemes based on AuthService
        if auth_service:
            for idp_client in auth_service.idp_clients:
                openapi_schema["components"]["securitySchemes"][
                    f"{idp_client.scheme_name}"
                ]["x-tokenName"] = idp_client.token_name

        return openapi_schema

    return partial(
        custom_openapi_function,
        _GET_OPEN_API_DEFAULTS,
        get_open_api_kwargs,
        fix_schema,
        auth_service,
    )


def fix_schema_nullable_and_single_element(schema: dict) -> None:
    """
    Fixes the schema by handling 'anyOf' constructs and setting the 'nullable' property.
    This function performs the following operations on the given schema:
    1. Replaces 'anyOf' constructs containing {"type": "null"} with 'nullable': True.
    2. Simplifies 'anyOf', 'allOf', and 'oneOf' constructs containing a single item by replacing them with that item.
    3. Recursively processes nested dictionaries and lists to apply the above transformations.
    Args:
        schema (dict): The JSON schema to be fixed.
    Returns:
        None: The function modifies the input schema in place.


    """
    # No other fix found online as of 2023-11-15.
    keys = list(schema.keys())
    for key in keys:
        if key in {"anyOf"}:
            # Search for presence of type: null in list
            for i, value in enumerate(schema[key]):
                has_null = value == {"type": "null"}
                if has_null:
                    break
            if not has_null:
                continue
            # Remove type: null from list
            del schema[key][i]  # pylint: disable=undefined-loop-variable

            # In case only 1 remaining item in list,
            # move it one level higher and remove anyOf key
            if len(schema[key]) == 1:
                key2 = list(schema[key][0].keys())[0]
                schema[key2] = schema[key][0][key2]
                del schema[key]
            # Set nullable property
            schema["nullable"] = True
        elif isinstance(schema[key], dict):
            # Continue recursively
            fix_schema_nullable_and_single_element(schema[key])
        elif isinstance(schema[key], list):
            # Continue recursively
            for item in schema[key]:
                if isinstance(item, dict):
                    fix_schema_nullable_and_single_element(item)
        if (
            key in schema
            and key in {"anyOf", "oneOf", "allOf"}
            and len(schema[key]) == 1
        ):
            # In case only 1 remaining item in list,
            # move it one level higher and remove list
            value = schema[key][0]
            del schema[key]
            for key2 in value:
                schema[key2] = value[key2]


# TODO: add a function to fix read-only fields



import json
import uuid

from pydantic import Field, field_serializer, field_validator

from gen_epix.seqdb.domain import enum


def str_uuid4() -> str:
    return str(uuid.uuid4())


class SeqMixin:
    seq: str = Field(
        description="The sequence in the representation defined by seq_format"
    )
    seq_format: enum.SeqFormat = Field(
        default=enum.SeqFormat.STR_DNA5,
        description="The format of the sequence",
    )
    seq_hash_sha256: bytes = Field(
        description="The SHA256 hash of the lower case ASCII encoded sequence without gaps. In the case of multiple contigs, this is the hash of the lexicographically sorted and concatenated contig sequences with a single gap ('-') as separator.",
        min_length=32,
        max_length=32,
    )
    length: int = Field(
        description="The length of the sequence. In case of multiple contigs, this is the sum of the lengths of all contigs.",
        ge=1,
    )

    @field_validator("seq_hash_sha256", mode="before")
    def _validate_seq_hash_sha256(cls, value: str | bytes) -> bytes:
        if isinstance(value, str):
            value = bytes.fromhex(value)
        return value

    @field_serializer("seq_format", mode="plain")
    def _serialize_seq_format(self, value: str | enum.SeqFormat) -> str:
        if isinstance(value, enum.SeqFormat):
            return value.value
        return value

    # TODO: adding the serializer gives issues writing as binary to the database, but not adding it may give other issues
    # @field_serializer("seq_hash_sha256", mode="plain")
    # def _serialize_seq_hash_sha256(self, value: str | bytes) -> str:
    #     if isinstance(value, bytes):
    #         return value.hex()
    #     return value


class CodeMixin:
    code: str = Field(
        default_factory=str_uuid4,
        description="A unique code for the instance, e.g. for external reference. Defaults to a UUID4.",
        max_length=255,
    )


class QualityMixin:
    quality_score: float | None = Field(
        default=None, description="The quality of the sequence, as a numerical value."
    )
    quality: enum.QualityControlResult | None = Field(
        default=None, description="The quality control result of the sequence."
    )

    @field_serializer("quality", mode="plain")
    def _serialize_quality(self, value: str | enum.QualityControlResult) -> str:
        if isinstance(value, enum.QualityControlResult):
            return value.value
        return value


class AlignmentMixin:
    aln: str = Field(
        description="The alignment in the representation defined by alignment_format"
    )
    aln_format: enum.AlignmentFormat = Field(
        default=enum.AlignmentFormat.CIGAR,
        description="The format of the alignment",
    )
    aln_hash_sha256: bytes = Field(
        description="The SHA256 hash of the ASCII lower case aligned reference sequence followed by the aligned contig seq.",
        min_length=32,
        max_length=32,
    )


class ProtocolMixin:
    code: str = Field(description="The code of the protocol", max_length=255)
    name: str = Field(description="The name of the protocol", max_length=255)
    version: str | None = Field(
        default=None, description="The version of the protocol", max_length=255
    )
    description: str | None = Field(
        default=None, description="The description of the protocol"
    )
    props: dict[str, str] = Field(
        default_factory=dict, description="The properties of the protocol"
    )

    @field_validator("props", mode="before")
    def _validate_props(cls, value: str | dict) -> dict:
        if isinstance(value, str):
            value = json.loads(value)
        return value



# pylint: disable=useless-import-alias
from gen_epix.casedb.domain.policy.abac import BaseCaseAbacPolicy as BaseCaseAbacPolicy
from gen_epix.casedb.domain.policy.permission import RoleGenerator as RoleGenerator
from gen_epix.commondb.domain.policy import (
    BaseIsOrganizationAdminPolicy as BaseIsOrganizationAdminPolicy,
)
from gen_epix.commondb.domain.policy import (
    BaseReadOrganizationResultsOnlyPolicy as BaseReadOrganizationResultsOnlyPolicy,
)
from gen_epix.commondb.domain.policy import (
    BaseReadSelfResultsOnlyPolicy as BaseReadSelfResultsOnlyPolicy,
)
from gen_epix.commondb.domain.policy import BaseUpdateUserPolicy as BaseUpdateUserPolicy



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from datetime import datetime
from typing import Any
from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped, relationship

from gen_epix.casedb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model import (
    RowMetadataMixin,
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.CASE.value)


class TreeAlgorithmClass(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.TreeAlgorithmClass)

    code: Mapped[str] = create_mapped_column(DOMAIN, model.TreeAlgorithmClass, "code")
    name: Mapped[str] = create_mapped_column(DOMAIN, model.TreeAlgorithmClass, "name")
    is_seq_based: Mapped[bool] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "is_seq_based"
    )
    is_dist_based: Mapped[bool] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "is_dist_based"
    )
    rank: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "rank"
    )


class TreeAlgorithm(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.TreeAlgorithm)

    tree_algorithm_class_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "tree_algorithm_class_id"
    )
    seqdb_tree_algorithm_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "seqdb_tree_algorithm_id"
    )
    code: Mapped[enum.TreeAlgorithmType] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "code"
    )
    name: Mapped[str] = create_mapped_column(DOMAIN, model.TreeAlgorithm, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "description"
    )
    is_ultrametric: Mapped[bool] = create_mapped_column(
        DOMAIN, model.TreeAlgorithm, "is_ultrametric"
    )

    tree_algorithm_class: Mapped[TreeAlgorithmClass] = relationship(
        TreeAlgorithmClass, foreign_keys=[tree_algorithm_class_id]
    )
    rank: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.TreeAlgorithmClass, "rank"
    )


class GeneticDistanceProtocol(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.GeneticDistanceProtocol)

    seqdb_seq_distance_protocol_id: Mapped[UUID] = create_mapped_column(
        DOMAIN,
        model.GeneticDistanceProtocol,
        "seqdb_seq_distance_protocol_id",
    )
    name: Mapped[str] = create_mapped_column(
        DOMAIN, model.GeneticDistanceProtocol, "name"
    )
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.GeneticDistanceProtocol, "description"
    )
    seqdb_max_stored_distance: Mapped[float | None] = create_mapped_column(
        DOMAIN,
        model.GeneticDistanceProtocol,
        "seqdb_max_stored_distance",
    )
    min_scale_unit: Mapped[float] = create_mapped_column(
        DOMAIN, model.GeneticDistanceProtocol, "min_scale_unit"
    )


class Dim(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Dim)

    dim_type: Mapped[enum.DimType] = create_mapped_column(DOMAIN, model.Dim, "dim_type")
    code: Mapped[str] = create_mapped_column(DOMAIN, model.Dim, "code")
    label: Mapped[str] = create_mapped_column(DOMAIN, model.Dim, "label")
    rank: Mapped[int | None] = create_mapped_column(DOMAIN, model.Dim, "rank")
    col_code_prefix: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Dim, "col_code_prefix"
    )
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Dim, "description"
    )
    props: Mapped[dict[str, Any]] = create_mapped_column(DOMAIN, model.Dim, "props")


class Col(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Col)

    dim_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.Col, "dim_id")
    code_suffix: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Col, "code_suffix"
    )
    code: Mapped[str] = create_mapped_column(DOMAIN, model.Col, "code")
    rank_in_dim: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.Col, "rank_in_dim"
    )
    label: Mapped[str | None] = create_mapped_column(DOMAIN, model.Col, "label")
    col_type: Mapped[enum.ColType] = create_mapped_column(DOMAIN, model.Col, "col_type")
    concept_set_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Col, "concept_set_id"
    )
    region_set_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Col, "region_set_id"
    )
    genetic_distance_protocol_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Col, "genetic_distance_protocol_id"
    )
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.Col, "description"
    )
    props: Mapped[dict[str, Any]] = create_mapped_column(DOMAIN, model.Col, "props")

    dim: Mapped[Dim] = relationship(Dim, foreign_keys=[dim_id])


class CaseType(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseType)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.CaseType, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseType, "description"
    )
    disease_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.CaseType, "disease_id"
    )
    etiological_agent_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.CaseType, "etiological_agent_id"
    )


class CaseTypeSetCategory(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseTypeSetCategory)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.CaseTypeSetCategory, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseTypeSetCategory, "description"
    )
    rank: Mapped[int] = create_mapped_column(DOMAIN, model.CaseTypeSetCategory, "rank")
    purpose: Mapped[enum.CaseTypeSetCategoryPurpose] = create_mapped_column(
        DOMAIN, model.CaseTypeSetCategory, "purpose"
    )


class CaseTypeSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseTypeSet)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.CaseTypeSet, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseTypeSet, "description"
    )
    case_type_set_category_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseTypeSet, "case_type_set_category_id"
    )
    rank: Mapped[float] = create_mapped_column(DOMAIN, model.CaseTypeSet, "rank")

    case_type_set_category: Mapped[CaseTypeSetCategory] = relationship(
        CaseTypeSetCategory, foreign_keys=[case_type_set_category_id]
    )


class CaseTypeSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseTypeSetMember)

    case_type_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseTypeSetMember, "case_type_set_id"
    )
    case_type_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseTypeSetMember, "case_type_id"
    )

    case_type_set: Mapped[CaseTypeSet] = relationship(
        CaseTypeSet, foreign_keys=[case_type_set_id]
    )
    case_type: Mapped[CaseType] = relationship(CaseType, foreign_keys=[case_type_id])


class CaseTypeCol(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseTypeCol)

    case_type_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "case_type_id"
    )
    col_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.CaseTypeCol, "col_id")
    occurrence: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "occurrence"
    )
    code: Mapped[str] = create_mapped_column(DOMAIN, model.CaseTypeCol, "code")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "description"
    )
    rank: Mapped[int | None] = create_mapped_column(DOMAIN, model.CaseTypeCol, "rank")
    label: Mapped[str | None] = create_mapped_column(DOMAIN, model.CaseTypeCol, "label")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "description"
    )
    min_value: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "min_value"
    )
    max_value: Mapped[float | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "max_value"
    )
    min_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "min_datetime"
    )
    max_datetime: Mapped[datetime | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "max_datetime"
    )
    min_length: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "min_length"
    )
    max_length: Mapped[int | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "max_length"
    )
    pattern: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "pattern"
    )
    ncbi_taxid: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "ncbi_taxid"
    )
    genetic_sequence_case_type_col_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "genetic_sequence_case_type_col_id"
    )
    tree_algorithm_codes: Mapped[list[enum.TreeAlgorithmType] | None] = (
        create_mapped_column(DOMAIN, model.CaseTypeCol, "tree_algorithm_codes")
    )
    props: Mapped[dict[str, Any]] = create_mapped_column(
        DOMAIN, model.CaseTypeCol, "props"
    )

    case_type: Mapped[CaseType] = relationship(CaseType, foreign_keys=[case_type_id])
    col: Mapped[Col] = relationship(Col, foreign_keys=[col_id])


class CaseTypeColSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseTypeColSet)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.CaseTypeColSet, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseTypeColSet, "description"
    )


class CaseTypeColSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseTypeColSetMember)

    case_type_col_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseTypeColSetMember, "case_type_col_set_id"
    )
    case_type_col_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseTypeColSetMember, "case_type_col_id"
    )

    case_type_col_set: Mapped[CaseTypeColSet] = relationship(
        CaseTypeColSet, foreign_keys=[case_type_col_set_id]
    )
    case_type_col: Mapped[CaseTypeCol] = relationship(
        CaseTypeCol, foreign_keys=[case_type_col_id]
    )


class Case(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Case)

    case_type_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.Case, "case_type_id"
    )
    subject_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Case, "subject_id"
    )
    created_in_data_collection_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.Case, "created_in_data_collection_id"
    )
    count: Mapped[int | None] = create_mapped_column(DOMAIN, model.Case, "count")
    case_date: Mapped[datetime] = create_mapped_column(DOMAIN, model.Case, "case_date")
    content: Mapped[dict[UUID, str]] = create_mapped_column(
        DOMAIN, model.Case, "content"
    )

    case_type: Mapped[CaseType] = relationship(CaseType, foreign_keys=[case_type_id])


class CaseDataCollectionLink(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseDataCollectionLink)

    case_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseDataCollectionLink, "case_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseDataCollectionLink, "data_collection_id"
    )

    case: Mapped[Case] = relationship(Case, foreign_keys=[case_id])


class CaseSetCategory(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseSetCategory)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.CaseSetCategory, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseSetCategory, "description"
    )


class CaseSetStatus(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseSetStatus)

    name: Mapped[str] = create_mapped_column(DOMAIN, model.CaseSetStatus, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.CaseSetStatus, "description"
    )


class CaseSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseSet)

    case_type_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseSet, "case_type_id"
    )
    created_in_data_collection_id: Mapped[UUID | None] = create_mapped_column(
        DOMAIN, model.CaseSet, "created_in_data_collection_id"
    )
    name: Mapped[str] = create_mapped_column(DOMAIN, model.CaseSet, "name")
    description: Mapped[str] = create_mapped_column(
        DOMAIN, model.CaseSet, "description"
    )
    created_at: Mapped[datetime] = create_mapped_column(
        DOMAIN, model.CaseSet, "created_at"
    )
    case_set_category_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseSet, "case_set_category_id"
    )
    case_set_status_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseSet, "case_set_status_id"
    )

    case_type: Mapped[CaseType] = relationship(CaseType, foreign_keys=[case_type_id])
    case_set_category: Mapped[CaseSetCategory] = relationship(
        CaseSetCategory, foreign_keys=[case_set_category_id]
    )
    case_set_status: Mapped[CaseSetStatus] = relationship(
        CaseSetStatus, foreign_keys=[case_set_status_id]
    )


class CaseSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseSetMember)

    case_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseSetMember, "case_set_id"
    )
    case_id: Mapped[UUID] = create_mapped_column(DOMAIN, model.CaseSetMember, "case_id")
    classification: Mapped[enum.CaseClassification] = create_mapped_column(
        DOMAIN, model.CaseSetMember, "classification"
    )

    case_set: Mapped[CaseSet] = relationship(CaseSet, foreign_keys=[case_set_id])
    case: Mapped[Case] = relationship(Case, foreign_keys=[case_id])


class CaseSetDataCollectionLink(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.CaseSetDataCollectionLink)

    case_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseSetDataCollectionLink, "case_set_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.CaseSetDataCollectionLink, "data_collection_id"
    )

    case_set: Mapped[CaseSet] = relationship(CaseSet, foreign_keys=[case_set_id])



import itertools
import json
from collections.abc import Callable, Hashable
from typing import Any
from uuid import UUID

from fastapi import APIRouter, FastAPI

from gen_epix.fastapp import exc, model
from gen_epix.fastapp.api import exc as api_exc
from gen_epix.fastapp.api.crud_endpoint_set import CrudEndpointSet
from gen_epix.fastapp.app import App
from gen_epix.fastapp.domain.entity import Entity
from gen_epix.fastapp.enum import (
    CrudEndpointType,
    CrudOperation,
    HttpMethod,
    PermissionType,
    PermissionTypeSet,
    StringCasing,
)
from gen_epix.fastapp.model import Permission
from gen_epix.filter import (
    CompositeFilter,
    Filter,
    TypedCompositeFilter,
    TypedDateRangeFilter,
    TypedDatetimeRangeFilter,
    TypedEqualsBooleanFilter,
    TypedEqualsNumberFilter,
    TypedEqualsStringFilter,
    TypedEqualsUuidFilter,
    TypedExistsFilter,
    TypedNoFilter,
    TypedNumberRangeFilter,
    TypedNumberSetFilter,
    TypedPartialDateRangeFilter,
    TypedRegexFilter,
    TypedStringSetFilter,
    TypedUuidSetFilter,
)


def _default_validate_query_filter(
    query_filter: Filter,
) -> bool:
    if isinstance(query_filter, CompositeFilter):
        for subfilter in query_filter.filters:
            if isinstance(subfilter, CompositeFilter):
                return False
    return True


class CrudEndpointGenerator:
    DEFAULT_BATCH_ROUTE_SUFFIX = "/batch"
    DEFAULT_QUERY_ROUTE_SUFFIX = "/query"
    DEFAULT_IDS_ROUTE_SUFFIX = "/ids"

    CRUD_OPERATION_TO_ENDPOINT_TYPE: dict[CrudOperation, CrudEndpointType] = {
        CrudOperation.READ_ALL: CrudEndpointType.GET_ALL,
        CrudOperation.READ_SOME: CrudEndpointType.GET_SOME,
        CrudOperation.READ_ONE: CrudEndpointType.GET_ONE,
        CrudOperation.CREATE_ONE: CrudEndpointType.POST_ONE,
        CrudOperation.CREATE_SOME: CrudEndpointType.POST_SOME,
        CrudOperation.UPDATE_ONE: CrudEndpointType.PUT_ONE,
        CrudOperation.UPDATE_SOME: CrudEndpointType.PUT_SOME,
        CrudOperation.DELETE_ONE: CrudEndpointType.DELETE_ONE,
        CrudOperation.DELETE_SOME: CrudEndpointType.DELETE_SOME,
        CrudOperation.DELETE_ALL: CrudEndpointType.DELETE_ALL,
    }
    CRUD_ENDPOINT_TYPE_ORDER: list[CrudEndpointType] = [
        CrudEndpointType.GET_ALL,
        CrudEndpointType.DELETE_ALL,
        CrudEndpointType.POST_QUERY,
        CrudEndpointType.POST_QUERY_IDS,
        CrudEndpointType.POST_SOME,
        CrudEndpointType.GET_SOME,
        CrudEndpointType.PUT_SOME,
        CrudEndpointType.DELETE_SOME,
        CrudEndpointType.POST_ONE,
        CrudEndpointType.GET_ONE,
        CrudEndpointType.PUT_ONE,
        CrudEndpointType.DELETE_ONE,
    ]

    @staticmethod
    def convert_ids_string_to_list(
        id_class: type, ids_str: str
    ) -> tuple[list | None, list[str]]:
        invalid_ids = []
        try:
            ids = [id_class(x) for x in ids_str.split(",")]
        except:
            try:
                raw_ids = json.loads(ids_str)
                ids = []
                for id_ in raw_ids:
                    try:
                        ids.append(id_class(id_))
                    except:
                        invalid_ids.append(id_)
            except:
                invalid_ids.append(ids_str)
                ids = None
        return ids, invalid_ids

    @staticmethod
    def generate_get_all(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
    ) -> None:
        async def endpoint_function(user: route.user_dependency) -> Any:  # type: ignore
            obj_ids = None
            cmd = route.crud_command_class(
                user=user,
                operation=CrudOperation.READ_ALL,
            )
            try:
                retval = route.app.handle(cmd)
                if route.model_class is not route.read_api_model_class:
                    retval = [route.read_api_model_class.from_model(x) for x in retval]
            except Exception as exception:
                handle_exception_fn(
                    "79d26f4f" + route.endpoint_basename,
                    user,
                    exception,
                    request_ids=obj_ids,
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename,
            endpoint_function,
            HttpMethod.GET,
            list[route.read_api_model_class],
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__get_all",
        )

    @staticmethod
    def generate_get_some(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
        batch_route_suffix: str | None = None,
    ) -> None:
        if not batch_route_suffix:
            batch_route_suffix = CrudEndpointGenerator.DEFAULT_BATCH_ROUTE_SUFFIX

        async def endpoint_function(user: route.user_dependency, ids: str) -> Any:  # type: ignore
            obj_ids, invalid_obj_ids = CrudEndpointGenerator.convert_ids_string_to_list(
                route.id_class, ids
            )
            if invalid_obj_ids:
                # f-string parsing fails if this is not first passed to a variable
                error_msg = ", ".join([f'"{x}"' for x in invalid_obj_ids])
                handle_exception_fn(
                    "a8abf0e3",
                    user,
                    exc.InvalidIdsError(
                        f"Invalid ids in ids query parameter: {error_msg}",
                        invalid_obj_ids,
                    ),
                    request_ids=invalid_obj_ids,
                )
            cmd = route.crud_command_class(
                user=user,
                obj_ids=obj_ids,
                operation=CrudOperation.READ_SOME,
            )
            try:
                retval = route.app.handle(cmd)
                if route.model_class is not route.read_api_model_class:
                    retval = [route.read_api_model_class.from_model(x) for x in retval]

            # TODO: Add a specific exception for NotImplementedError
            except Exception as exception:
                handle_exception_fn(
                    "fbc53f5d" + route.endpoint_basename,
                    user,
                    exception,
                    request_ids=obj_ids,
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + batch_route_suffix,
            endpoint_function,
            HttpMethod.GET,
            list[route.read_api_model_class],
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__get_some",
        )

    @staticmethod
    def generate_post_query(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
        query_route_suffix: str | None = None,
        return_id: bool = False,
        ids_route_suffix: str | None = None,
        validate_query_filter: (
            Callable[[Filter], bool] | None
        ) = _default_validate_query_filter,
    ) -> None:
        if not query_route_suffix:
            query_route_suffix = CrudEndpointGenerator.DEFAULT_QUERY_ROUTE_SUFFIX
        if not ids_route_suffix:
            ids_route_suffix = CrudEndpointGenerator.DEFAULT_IDS_ROUTE_SUFFIX
        route_suffix = query_route_suffix
        if return_id:
            route_suffix = route_suffix + ids_route_suffix
        operation_id = route.operation_id_basename or route.endpoint_basename
        operation_id += "__post_query"
        if return_id:
            operation_id += "__ids"

        async def endpoint_function(
            user: route.user_dependency,  # type: ignore
            filter: (
                TypedExistsFilter
                | TypedEqualsBooleanFilter
                | TypedEqualsNumberFilter
                | TypedEqualsStringFilter
                | TypedEqualsUuidFilter
                | TypedNumberRangeFilter
                | TypedDateRangeFilter
                | TypedDatetimeRangeFilter
                | TypedPartialDateRangeFilter
                | TypedRegexFilter
                | TypedNumberSetFilter
                | TypedStringSetFilter
                | TypedUuidSetFilter
                | TypedNoFilter
                | TypedCompositeFilter
            ),
        ) -> Any:
            if validate_query_filter and not validate_query_filter(filter):
                handle_exception_fn(
                    "cee23041" + route.endpoint_basename,
                    user,
                    exc.InvalidArgumentsError("Invalid filter"),
                )
            cmd = route.crud_command_class(
                user=user,
                operation=CrudOperation.READ_ALL,
                query_filter=filter,
                props={
                    "return_id": return_id,
                },
            )
            try:
                retval = route.app.handle(cmd)
                if (
                    not return_id
                    and route.model_class is not route.read_api_model_class
                ):
                    retval = [route.read_api_model_class.from_model(x) for x in retval]

            # TODO: Add a specific exception for NotImplementedError
            except Exception as exception:
                handle_exception_fn(
                    "ca2591fa" + route.endpoint_basename, user, exception
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + route_suffix,
            endpoint_function,
            HttpMethod.POST,
            list[route.id_class] if return_id else list[route.read_api_model_class],
            route,
            operation_id=operation_id,
        )

    @staticmethod
    def generate_get_one(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
    ) -> None:
        async def endpoint_function(
            user: route.user_dependency,  # type: ignore
            object_id: route.id_class,  # type: ignore
        ) -> Any:
            try:
                cmd = route.crud_command_class(
                    user=user,
                    operation=CrudOperation.READ_ONE,
                    obj_ids=object_id,
                )
                obj = route.app.handle(cmd)
                if route.model_class is not route.read_api_model_class:
                    obj = route.read_api_model_class.from_model(obj)

            # TODO: Add a specific exception for NotImplementedError
            except Exception as exception:
                handle_exception_fn(
                    "3680417c" + route.endpoint_basename + f"/{object_id}",
                    user,
                    exception,
                    request_ids=[object_id],
                )
            return obj

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + "/{object_id}",
            endpoint_function,
            HttpMethod.GET,
            route.read_api_model_class,
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__get_one",
        )

    @staticmethod
    def generate_post_one(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
    ) -> None:
        async def endpoint_function(
            user: route.user_dependency, create_obj: route.create_api_model_class  # type: ignore
        ) -> Any:
            try:
                cmd = route.crud_command_class(
                    user=user,
                    operation=CrudOperation.CREATE_ONE,
                    objs=(
                        create_obj
                        if route.model_class is route.create_api_model_class
                        else route.create_api_model_class.to_model(create_obj)
                    ),
                    props={"return_id": route.post_returns_id},
                )
                retval = route.app.handle(cmd)
                if (
                    not route.post_returns_id
                    and route.model_class is not route.read_api_model_class
                ):
                    retval = route.read_api_model_class.from_model(retval)

            except Exception as exception:
                try:
                    request_ids = [create_obj.id]
                except:
                    request_ids = None
                handle_exception_fn(
                    "02c70ca7" + route.endpoint_basename,
                    user,
                    exception,
                    request_ids=request_ids,
                )

            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename,
            endpoint_function,
            HttpMethod.POST,
            route.id_class if route.post_returns_id else route.read_api_model_class,
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__post_one",
        )

    @staticmethod
    def generate_post_some(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
        batch_route_suffix: str | None = None,
    ) -> None:
        if not batch_route_suffix:
            batch_route_suffix = CrudEndpointGenerator.DEFAULT_BATCH_ROUTE_SUFFIX

        async def endpoint_function(
            user: route.user_dependency, create_objs: list[route.create_api_model_class]  # type: ignore
        ) -> Any:
            try:
                cmd = route.crud_command_class(
                    user=user,
                    operation=CrudOperation.CREATE_SOME,
                    objs=(
                        create_objs
                        if route.model_class is route.create_api_model_class
                        else [
                            route.create_api_model_class.to_model(x)
                            for x in create_objs
                        ]
                    ),
                    props={"return_id": route.post_returns_id},
                )
                retval = route.app.handle(cmd)
                if (
                    not route.post_returns_id
                    and route.model_class is not route.read_api_model_class
                ):
                    retval = [route.read_api_model_class.from_model(x) for x in retval]

            except Exception as exception:
                try:
                    request_ids = [x.id for x in create_objs]
                except:
                    request_ids = None
                handle_exception_fn(
                    "e96480ac" + route.endpoint_basename,
                    user,
                    exception,
                    request_ids=request_ids,
                )

            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + batch_route_suffix,
            endpoint_function,
            HttpMethod.POST,
            (
                list[route.id_class]
                if route.post_returns_id
                else list[route.read_api_model_class]
            ),
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__post_some",
        )

    @staticmethod
    def generate_put_one(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
    ) -> None:
        async def endpoint_function(
            user: route.user_dependency,
            object_id: route.id_class,  # type: ignore
            update_obj: route.create_api_model_class,  # type: ignore
        ) -> Any:
            if update_obj.id != object_id:
                raise api_exc.BadRequest400HTTPException()
            try:
                cmd = route.crud_command_class(
                    user=user,
                    operation=CrudOperation.UPDATE_ONE,
                    objs=(
                        update_obj
                        if route.model_class is route.create_api_model_class
                        else route.model_class.to_model(update_obj)
                    ),
                    props={"return_id": route.put_returns_id},
                )
                retval = route.app.handle(cmd)
                if (
                    not route.put_returns_id
                    and route.model_class is not route.read_api_model_class
                ):
                    retval = route.read_api_model_class.from_model(retval)
            except Exception as exception:
                handle_exception_fn(
                    "1459d302" + route.endpoint_basename + f"/{object_id}",
                    user,
                    exception,
                    request_ids=[object_id],
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + "/{object_id}",
            endpoint_function,
            HttpMethod.PUT,
            route.id_class if route.post_returns_id else route.read_api_model_class,
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__put_one",
        )

    @staticmethod
    def generate_put_some(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
        batch_route_suffix: str | None = None,
    ) -> None:
        if not batch_route_suffix:
            batch_route_suffix = CrudEndpointGenerator.DEFAULT_BATCH_ROUTE_SUFFIX

        async def endpoint_function(
            user: route.user_dependency,  # type: ignore
            update_objs: list[route.create_api_model_class],  # type: ignore
        ) -> Any:
            try:
                cmd = route.crud_command_class(
                    user=user,
                    operation=CrudOperation.UPDATE_SOME,
                    objs=(
                        update_objs
                        if route.model_class is route.create_api_model_class
                        else [route.model_class.to_model(x) for x in update_objs]
                    ),
                    props={"return_id": route.put_returns_id},
                )
                retval = route.app.handle(cmd)
                if (
                    not route.put_returns_id
                    and route.model_class is not route.read_api_model_class
                ):
                    retval = [route.read_api_model_class.from_model(x) for x in retval]
            except Exception as exception:
                handle_exception_fn(
                    "b9359ae3" + route.endpoint_basename,
                    user,
                    exception,
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + batch_route_suffix,
            endpoint_function,
            HttpMethod.PUT,
            (
                list[route.id_class]
                if route.post_returns_id
                else list[route.read_api_model_class]
            ),
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__put_some",
        )

    @staticmethod
    def generate_delete_one(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
    ) -> None:
        async def endpoint_function(user: route.user_dependency, object_id: Any) -> Any:  # type: ignore
            # TODO: distinguish between soft and hard delete through hard_delete:
            #  bool = False parameter
            try:
                cmd = route.crud_command_class(
                    user=user,
                    operation=CrudOperation.DELETE_ONE,
                    obj_ids=object_id,
                )
                retval = route.app.handle(cmd)
            except Exception as exception:
                handle_exception_fn(
                    "ab4df15f" + route.endpoint_basename + f"/{object_id}",
                    user,
                    exception,
                    request_ids=[object_id],
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + "/{object_id}",
            endpoint_function,
            HttpMethod.DELETE,
            route.id_class,
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__delete_one",
        )

    @staticmethod
    def generate_delete_all(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
    ) -> None:
        async def endpoint_function(user: route.user_dependency) -> Any:  # type: ignore
            obj_ids = None
            cmd = route.crud_command_class(
                user=user,
                operation=CrudOperation.DELETE_ALL,
                props={"return_id": route.delete_all_returns_id},
            )
            try:
                retval = route.app.handle(cmd)
            # TODO: Add a specific exception for NotImplementedError
            except Exception as exception:
                handle_exception_fn(
                    "79d26f4f" + route.endpoint_basename,
                    user,
                    exception,
                    request_ids=obj_ids,
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename,
            endpoint_function,
            HttpMethod.DELETE,
            list[route.id_class] if route.delete_all_returns_id else None,
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__delete_all",
        )

    @staticmethod
    def generate_delete_some(
        fast_api: FastAPI | APIRouter,
        route: CrudEndpointSet,
        handle_exception_fn: Callable,
        batch_route_suffix: str | None = None,
    ) -> None:
        if not batch_route_suffix:
            batch_route_suffix = CrudEndpointGenerator.DEFAULT_BATCH_ROUTE_SUFFIX

        async def endpoint_function(
            user: route.user_dependency,  # type: ignore
            ids: str,
        ) -> Any:
            obj_ids, invalid_obj_ids = CrudEndpointGenerator.convert_ids_string_to_list(
                route.id_class, ids
            )
            if invalid_obj_ids:
                # f-string parsing fails if this is not first passed to a variable
                error_msg = ", ".join([f'"{x}"' for x in invalid_obj_ids])
                handle_exception_fn(
                    "e73f930b",
                    user,
                    exc.InvalidIdsError(
                        f"Invalid ids in ids query parameter: {error_msg}",
                        invalid_obj_ids,
                    ),
                    request_ids=invalid_obj_ids,
                )
            cmd = route.crud_command_class(
                user=user,
                obj_ids=obj_ids,
                operation=CrudOperation.DELETE_SOME,
                props={"return_id": route.delete_all_returns_id},
            )
            try:
                retval = route.app.handle(cmd)
            # TODO: Add a specific exception for NotImplementedError
            except Exception as exception:
                handle_exception_fn(
                    "fce475e1" + route.endpoint_basename,
                    user,
                    exception,
                    request_ids=obj_ids,
                )
            return retval

        CrudEndpointGenerator._add_route(
            fast_api,
            route.endpoint_basename + batch_route_suffix,
            endpoint_function,
            HttpMethod.DELETE,
            list[route.id_class],
            route,
            operation_id=(route.operation_id_basename or route.endpoint_basename)
            + "__delete_some",
        )

    @staticmethod
    def _add_route(
        fast_api: FastAPI | APIRouter,
        endpoint: str,
        endpoint_fn: Callable,
        method: HttpMethod,
        response_model: Any | None,
        route: CrudEndpointSet,
        operation_id: str | None = None,
    ) -> None:
        if not operation_id:
            tokens = endpoint.split("/")
            if tokens[-1] == "{object_id}" or method.value.upper() == "POST":
                operation_id = tokens[1] + "__" + method.value.lower() + "_one"
            else:
                operation_id = tokens[1] + "__" + method.value.lower() + "_all"

        fast_api.add_api_route(
            "/" + endpoint,
            endpoint_fn,
            methods=[method.value],
            response_model=response_model,
            name=operation_id,
            description=route.description,
            operation_id=operation_id,
            response_model_exclude_none=bool(route.response_model_exclude_none),
        )

    @staticmethod
    def generate_endpoints(
        fast_api: FastAPI | APIRouter,
        routes: list[CrudEndpointSet],
        handle_exception_fn: Callable,
        batch_route_suffix: str | None = None,
        query_route_suffix: str | None = None,
        ids_route_suffix: str | None = None,
        validate_query_filter: (
            Callable[[Filter], bool] | None
        ) = _default_validate_query_filter,
    ) -> None:
        # Map endpoint types to functions
        function_map = {
            CrudEndpointType.GET_ALL: CrudEndpointGenerator.generate_get_all,
            CrudEndpointType.GET_SOME: CrudEndpointGenerator.generate_get_some,
            CrudEndpointType.POST_QUERY: CrudEndpointGenerator.generate_post_query,
            CrudEndpointType.POST_QUERY_IDS: CrudEndpointGenerator.generate_post_query,
            CrudEndpointType.GET_ONE: CrudEndpointGenerator.generate_get_one,
            CrudEndpointType.POST_ONE: CrudEndpointGenerator.generate_post_one,
            CrudEndpointType.POST_SOME: CrudEndpointGenerator.generate_post_some,
            CrudEndpointType.PUT_ONE: CrudEndpointGenerator.generate_put_one,
            CrudEndpointType.PUT_SOME: CrudEndpointGenerator.generate_put_some,
            CrudEndpointType.DELETE_ALL: CrudEndpointGenerator.generate_delete_all,
            CrudEndpointType.DELETE_SOME: CrudEndpointGenerator.generate_delete_some,
            CrudEndpointType.DELETE_ONE: CrudEndpointGenerator.generate_delete_one,
        }
        # Go over each route and create the endpoints
        for route in routes:
            # Create each endpoint type in order to avoid path conflicts, e.g. if
            # /batch overlaps with /{object_id} then /batch should be created first
            for endpoint_type in CrudEndpointGenerator.CRUD_ENDPOINT_TYPE_ORDER:
                if endpoint_type not in route.endpoint_types:
                    continue
                extra_args: dict[str, Any] = {}
                if endpoint_type == CrudEndpointType.POST_QUERY:
                    extra_args["query_route_suffix"] = query_route_suffix
                    extra_args["return_id"] = False
                    extra_args["ids_route_suffix"] = ids_route_suffix
                    extra_args["validate_query_filter"] = validate_query_filter
                elif endpoint_type == CrudEndpointType.POST_QUERY_IDS:
                    extra_args["query_route_suffix"] = query_route_suffix
                    extra_args["return_id"] = True
                    extra_args["ids_route_suffix"] = ids_route_suffix
                    extra_args["validate_query_filter"] = validate_query_filter
                elif endpoint_type == CrudEndpointType.POST_SOME:
                    extra_args["batch_route_suffix"] = batch_route_suffix
                elif endpoint_type == CrudEndpointType.GET_SOME:
                    extra_args["batch_route_suffix"] = batch_route_suffix
                elif endpoint_type == CrudEndpointType.PUT_SOME:
                    extra_args["batch_route_suffix"] = batch_route_suffix
                elif endpoint_type == CrudEndpointType.DELETE_SOME:
                    extra_args["batch_route_suffix"] = batch_route_suffix
                function_map[endpoint_type](  # type: ignore
                    fast_api, route, handle_exception_fn, **extra_args
                )

    @staticmethod
    def create_crud_endpoint_set_for_domain(
        app: App,
        service_type: Hashable | set[Hashable] | None = None,
        user_dependency: Callable | None = None,
        excluded_permissions: (
            set[Permission] | dict[type[model.Model], PermissionTypeSet | None] | None
        ) = None,
        excluded_crud_operations: (
            dict[type[model.Model], set[CrudOperation]] | None
        ) = None,
        excluded_crud_endpoint_types: (
            dict[type[model.Model], set[CrudEndpointType]] | None
        ) = None,
        default_description: str | None = None,
        endpoint_string_casing: StringCasing = StringCasing.SNAKE_CASE,
        query_filter_validator: (
            Callable[[Filter], bool] | None
        ) = _default_validate_query_filter,
    ) -> list[CrudEndpointSet]:
        # Parse exclusions
        if excluded_permissions is None:
            excluded_permissions = app.domain.get_model_excluded_permissions()  # type: ignore[assignment] # Unclear why raised
        parsed_excluded_permissions: set[Permission] = set()
        if isinstance(excluded_permissions, dict):
            parsed_excluded_permissions = set()
            for model_class, permission_type_set in excluded_permissions.items():
                parsed_excluded_permissions.update(
                    app.domain.get_permissions_for_model(
                        model_class, permission_type_set=permission_type_set
                    )
                )
        elif isinstance(excluded_permissions, set):
            parsed_excluded_permissions = excluded_permissions
        if excluded_crud_operations is None:
            excluded_crud_operations = {}
        if excluded_crud_endpoint_types is None:
            excluded_crud_endpoint_types = {}
        # Create CRUD endpoint sets
        if service_type is None:
            entities = app.domain.get_dag_sorted_entities()
        elif isinstance(service_type, Hashable):
            entities = app.domain.get_dag_sorted_entities(service_type=service_type)
        elif isinstance(service_type, set):
            entities = list(
                itertools.chain(
                    app.domain.get_dag_sorted_entities(service_type=x)
                    for x in service_type
                )
            )
        else:
            raise exc.DomainException(
                f"Invalid service type {service_type} for CRUD endpoint generation"
            )
        crud_endpoint_sets = []
        for entity in entities:
            if not entity.persistable:
                continue
            assert issubclass(entity.model_class, model.Model)
            crud_endpoint_sets.append(
                CrudEndpointGenerator.get_crud_endpoint_set_for_entity(
                    entity,
                    app,
                    user_dependency=user_dependency,
                    excluded_permissions=parsed_excluded_permissions,
                    excluded_crud_operations=excluded_crud_operations.get(
                        entity.model_class
                    ),
                    excluded_crud_endpoint_types=excluded_crud_endpoint_types.get(
                        entity.model_class
                    ),
                    default_description=default_description,
                    endpoint_string_casing=endpoint_string_casing,
                    query_filter_validator=query_filter_validator,
                )
            )
        return crud_endpoint_sets

    @staticmethod
    def get_crud_endpoint_set_for_entity(
        entity: Entity,
        app: App,
        user_dependency: Callable | None = None,
        add_query_route: bool = True,
        excluded_permissions: set[Permission] | None = None,
        excluded_crud_operations: set[CrudOperation] | None = None,
        excluded_crud_endpoint_types: set[CrudEndpointType] | None = None,
        default_description: str | None = None,
        endpoint_string_casing: StringCasing = StringCasing.SNAKE_CASE,
        endpoint_string_is_plural: bool = True,
        query_filter_validator: (
            Callable[[Filter], bool] | None
        ) = _default_validate_query_filter,
    ) -> CrudEndpointSet:
        # Initialize some
        model_class = entity.model_class
        assert issubclass(model_class, model.Model)
        if excluded_crud_endpoint_types is None:
            excluded_crud_endpoint_types = set()
        if excluded_crud_operations is None:
            excluded_crud_operations = set()
        if excluded_permissions is None:
            excluded_permissions = set()
        # Get endpoint types to create, starting from all permissions for the model and removing any excluded ones
        crud_command_class = entity.crud_command_class
        if crud_command_class is None:
            raise exc.DomainException(
                f"Entity {entity.name} does not have a crud command class"
            )
        assert issubclass(crud_command_class, model.CrudCommand)
        permissions = (
            app.domain.get_permissions_for_model(model_class) - excluded_permissions
        )
        crud_operations = (
            CrudEndpointGenerator.get_crud_operations_for_permissions(permissions)
            - excluded_crud_operations
        )
        crud_endpoint_types = (
            CrudEndpointGenerator.get_crud_endpoint_types_for_operations(
                crud_operations, add_query_route=add_query_route
            )
        ) - excluded_crud_endpoint_types
        # Create CRUD endpoint set
        endpoint_basename = CrudEndpointGenerator.get_endpoint_basename(
            entity, endpoint_string_casing, endpoint_string_is_plural
        )
        crud_endpoint_set = CrudEndpointSet(
            model_class=model_class,
            create_api_model_class=entity.create_api_model_class,
            read_api_model_class=entity.read_api_model_class,
            endpoint_basename=endpoint_basename,
            crud_command_class=crud_command_class,
            endpoint_types=crud_endpoint_types,
            operation_id_basename=entity.name,
            description=crud_command_class.__doc__ or default_description,
            # description=model_class.model_json_schema().get(
            #     "description", default_description
            # ),
            user_dependency=user_dependency,
            app=app,
            id_class=UUID,
            response_model_exclude_none=True,
            query_filter_validator=query_filter_validator,
        )
        return crud_endpoint_set

    @staticmethod
    def get_crud_operations_for_permissions(
        permissions: set[Permission],
    ) -> set[CrudOperation]:
        crud_operations = set()
        for permission in permissions:
            permission_type = permission.permission_type
            if permission_type == PermissionType.CREATE:
                crud_operations.update(
                    {CrudOperation.CREATE_ONE, CrudOperation.CREATE_SOME}
                )
            elif permission_type == PermissionType.READ:
                crud_operations.update(
                    {
                        CrudOperation.READ_ALL,
                        CrudOperation.READ_SOME,
                        CrudOperation.READ_ONE,
                    }
                )
            elif permission_type == PermissionType.UPDATE:
                crud_operations.update(
                    {CrudOperation.UPDATE_ONE, CrudOperation.UPDATE_SOME}
                )
            elif permission_type == PermissionType.DELETE:
                crud_operations.update(
                    {
                        CrudOperation.DELETE_ALL,
                        CrudOperation.DELETE_SOME,
                        CrudOperation.DELETE_ONE,
                    }
                )
            else:
                raise NotImplementedError(
                    f"Permission type {permission_type} not implemented"
                )
        return crud_operations

    @staticmethod
    def get_endpoint_basename(
        entity: Entity,
        string_casing: StringCasing = StringCasing.SNAKE_CASE,
        is_plural: bool = True,
    ) -> str:
        name = entity.get_name_by_casing(string_casing, is_plural=is_plural)
        if name is None:
            raise exc.DomainException(
                f"Entity {entity.name} does not have a {string_casing.value} name"
            )
        return name

    @staticmethod
    def get_crud_endpoint_types_for_operations(
        crud_operations: set[CrudOperation], add_query_route: bool = True
    ) -> set[CrudEndpointType]:
        crud_endpoint_types = set()
        for crud_operation in crud_operations:
            crud_endpoint_types.add(
                CrudEndpointGenerator.CRUD_OPERATION_TO_ENDPOINT_TYPE[crud_operation]
            )
            if crud_operation == CrudOperation.READ_ALL and add_query_route:
                crud_endpoint_types.update(
                    {
                        CrudEndpointType.POST_QUERY,
                        CrudEndpointType.POST_QUERY_IDS,
                    }
                )
        return crud_endpoint_types



from uuid import UUID

from pydantic import Field

from gen_epix.commondb.domain.command import Command

# Non-CRUD


class RetrieveGeneticSequenceByIdCommand(Command):
    """
    Retrieve a genetic sequence by its ID.
    """

    seq_ids: list[UUID] = Field(
        description="The IDs of the genetic sequences to retrieve."
    )


class RetrieveGeneticSequenceFastaByIdCommand(Command):
    """
    Retrieve a set of genetic sequences in FASTA format based on a set of sequence IDs.
    An iterator is returned that yields the FASTA lines.
    """

    seq_ids: list[UUID] = Field(
        description="The IDs of the genetic sequences to retrieve."
    )
    wrap: int = Field(
        default=80,
        description="The line length to wrap sequences at, or 0 for no wrapping.",
    )


# CRUD



# pylint: disable=useless-import-alias
from gen_epix.fastapp.app import App as App
from gen_epix.fastapp.domain import Domain as Domain
from gen_epix.fastapp.domain import Entity as Entity
from gen_epix.fastapp.domain import Key as Key
from gen_epix.fastapp.domain import Link as Link
from gen_epix.fastapp.enum import AuthProtocol as AuthProtocol
from gen_epix.fastapp.enum import CrudEndpointType as CrudEndpointType
from gen_epix.fastapp.enum import CrudOperation as CrudOperation
from gen_epix.fastapp.enum import CrudOperationSet as CrudOperationSet
from gen_epix.fastapp.enum import EventTiming as EventTiming
from gen_epix.fastapp.enum import FieldType as FieldType
from gen_epix.fastapp.enum import FieldTypeSet as FieldTypeSet
from gen_epix.fastapp.enum import HttpMethod as HttpMethod
from gen_epix.fastapp.enum import HttpProtocol as HttpProtocol
from gen_epix.fastapp.enum import IsolationLevel as IsolationLevel
from gen_epix.fastapp.enum import LogLevel as LogLevel
from gen_epix.fastapp.enum import OAuthFlow as OAuthFlow
from gen_epix.fastapp.enum import PermissionType as PermissionType
from gen_epix.fastapp.enum import PermissionTypeSet as PermissionTypeSet
from gen_epix.fastapp.enum import SortOrder as SortOrder
from gen_epix.fastapp.log import LogItem as LogItem
from gen_epix.fastapp.model import Command as Command
from gen_epix.fastapp.model import CrudCommand as CrudCommand
from gen_epix.fastapp.model import Model as Model
from gen_epix.fastapp.model import Permission as Permission
from gen_epix.fastapp.model import Policy as Policy
from gen_epix.fastapp.model import Role as Role
from gen_epix.fastapp.model import UpdateAssociationCommand as UpdateAssociationCommand
from gen_epix.fastapp.model import User as User
from gen_epix.fastapp.pdp import PolicyDecisionPoint as PolicyDecisionPoint
from gen_epix.fastapp.remote_app import RemoteApp as RemoteApp
from gen_epix.fastapp.repository import BaseRepository as BaseRepository
from gen_epix.fastapp.service import BaseService as BaseService
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork as BaseUnitOfWork
from gen_epix.fastapp.user_manager import BaseUserManager as BaseUserManager
